image *font;
int charx;
int chary;
char cstart[ 256 ];
char csize[ 256 ];
int cptr[ 256 ];
int xchars;
int ychars;
int empty = 2;

uchar *txt_ptr;
int linesize; //in pixels;
int lines; //number of lines;
int page_size = 0; //number of lines on the screen (in pixels);
long line_ptr[ 32000 ];
long line_size[ 32000 ];

int xoffset = 0;
float yoffset = 0;
float y_speed = 0; //pixels per second
float y_fade_speed = 60; //fade per second

COLOR text_color = COLORMASK;
COLOR black_text = COLORMASK;
COLOR white_text = COLORMASK;

int draw_page_num = 1;
int y_dec = 1;

int font_shadow = 0;
int font_bold = 0;
int font_select_y = -1;

int selected_line = -1;

int text_center = 0;

#define Y_LINE_SIZE ( chary - y_dec )

void load_font( char *filename, int xc, int yc )
{
    xchars = xc;
    ychars = yc;
    font = load_jpeg( filename, 1 );
    charx = ( font->xsize << 9 ) / xc;
    chary = ( font->ysize << 9 ) / yc;
    int c = 0;
    for( int yy = 0, ycc = 0; ycc < yc; yy += chary, ycc++ )
    {
	for( int xx = 0, xcc = 0; xcc < xc; xx += charx, xcc++ )
	{
	    int min_x = charx >> 9;
	    int max_x = 0;
	    int cp = ( yy >> 9 ) * font->xsize + ( xx >> 9 );
	    for( int y = 0; y < chary >> 9; y++ )
	    {
		for( int x = 0; x < charx >> 9; x++ )
		{
		    if( cp < font->xsize * font->ysize )
		    if( font->data[ cp ] > 5 )
		    {
			if( x < min_x ) min_x = x;
			if( x > max_x ) max_x = x;
		    }
		    cp++;
		}
		cp += font->xsize - ( charx >> 9 );
	    }
	    cstart[ c ] = (char)min_x;
	    csize[ c ] = (char)( max_x - min_x + 1 );
	    if( csize[ c ] < 0 ) { csize[ c ] = charx; cstart[ c ] = 0; }
	    cptr[ c ] = ( yy >> 9 ) * font->xsize + ( xx >> 9 ) + min_x;
	    c++;
	    if( c >= 256 ) break;
	}
	if( c >= 256 ) break;
    }
    charx >>= 9;
    chary >>= 9;
}

void close_font()
{
    clear_image( font );
}

int get_char_size( uchar c )
{
    if( c == ' ' ) return charx / 2;
    if( c < 32 ) return 0;
    if( c == '^' ) return 0;
    if( c == '#' ) return 0;
    if( c == 0x85 ) return csize[ '.' ] * 3 + ( empty * 2 );
    return csize[ c ];
}

int get_string_size( char *str, int offset, int size, window_manager *wm )
{
    int ss = 0;
    int c = offset;
    for( ; c < offset + size ; )
    {
	if( str[ c ] == 0 ) break;
	int cs = get_char_size( (uchar)str[ c ] );
	if( cs ) cs += empty;
	ss += cs;
	c++;
    }
    return ss;
}

int grayscale_img_draw_char( int x, int y, uchar c, image *img, window_manager *wm )
{
    if( c == ' ' ) return charx / 2;
    if( c < 32 ) return 0;
    if( 1 )
    {
	int p = cptr[ c ];
	int sp = y * img->xsize + x;
	int max = font->xsize * font->ysize;
	for( int yy = 0; yy < chary && p < max; yy++ )
	{
	    for( int xx = 0; xx < csize[ c ] && p < max; xx++ )
	    {
		if( font->data[ p ] ) 
		{
		    img->data[ sp ] = font->data[ p ];
		}
		p++;
		sp++;
	    }
	    p += font->xsize - csize[ c ];
	    sp += img->xsize - csize[ c ];
	}
    }
    return csize[ c ];
}

int screen_draw_char( int x, int y, uchar c, window_manager *wm )
{
    if( c == ' ' ) return charx / 2;
    if( c < 32 ) return 0;
    if( c == '^' )
    {
	//Special symbol:
	int ptr = ( y + ( chary / 2 ) ) * wm->pscreen_x + x;
	COLOR new_text_color = get_color( 80,80,80 );
	for( int a = 0; a < linesize; a += 2 )
	{
	    wm->g_screen[ ptr ] = new_text_color;
	    //wm->g_screen[ ptr + 1 ] = new_text_color;
	    //wm->g_screen[ ptr + wm->pscreen_x * 2 ] = new_text_color;
	    //wm->g_screen[ ptr + wm->pscreen_x * 2 + 1 ] = new_text_color;
	    ptr += 2;
	}
	return 0;
    }
    if( c == '#' )
    {
	//Special symbol:
	font_bold = 1;
	if( text_center )
	{
	    int bar_size = wm->pscreen_x / 3;
	    int ptr = ( y + 1 ) * wm->pscreen_x + ( wm->pscreen_x - bar_size ) / 2;
	    int xx = 0;
	    for( int a = 0; a < bar_size; a++ )
	    {
		int sval = 0;
		if( xx < bar_size / 2 )
		{
		    sval = ( xx << 11 ) / ( bar_size / 2 );
		}
		else
		{
		    sval = ( ( bar_size - xx ) << 11 ) / ( bar_size / 2 );
		}
		sval *= 255;
		sval >>= 11;
		COLOR cval = COLORMASK;
		if( wm->cur_transparency < 255 )
		{
		    sval *= wm->cur_transparency;
		    sval >>= 8;
		}
		//if( a & 1 ) cval = COLORMASK;
		for( int i = 0; i < chary - 2; i++ )
		{
		    wm->g_screen[ ptr + i * wm->pscreen_x ] = blend( wm->g_screen[ ptr + i * wm->pscreen_x ], cval, sval );
		}
		xx++;
		ptr++;
	    }
	}
	else
	{
	    int ptr = ( y + 1 ) * wm->pscreen_x + x - 1;
	    int xx = 0;
    	    int transp = 100;
	    if( wm->cur_transparency < 255 )
	    {
	        transp *= wm->cur_transparency;
	        transp >>= 8;
	    }
	    for( int a = 0; a < linesize + 1; a++ )
	    {
		for( int i = 0; i < chary - 2; i++ )
		{
		    wm->g_screen[ ptr + i * wm->pscreen_x ] = blend( wm->g_screen[ ptr + i * wm->pscreen_x ], text_color ^ COLORMASK, transp );
		}
		ptr++;
	    }
	    ptr = ( y + 5 ) * wm->pscreen_x + x + linesize - 8;
	    wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 1 ] = blend( wm->g_screen[ ptr + 1 ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 2 ] = blend( wm->g_screen[ ptr + 2 ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 1 + wm->pscreen_x ] = blend( wm->g_screen[ ptr + 1 + wm->pscreen_x ], 0, wm->cur_transparency );
	    ptr = ( y + 8 ) * wm->pscreen_x + x + linesize - 8;
	    wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 1 ] = blend( wm->g_screen[ ptr + 1 ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 2 ] = blend( wm->g_screen[ ptr + 2 ], 0, wm->cur_transparency );
	    wm->g_screen[ ptr + 1 + wm->pscreen_x ] = blend( wm->g_screen[ ptr + 1 + wm->pscreen_x ], 0, wm->cur_transparency );
	}
	return 0;
    }
    if( c == 0x85 ) 
    {
	screen_draw_char( x, y, '.', wm ); x += csize [ '.' ] + empty;
	screen_draw_char( x, y, '.', wm ); x += csize [ '.' ] + empty;
	screen_draw_char( x, y, '.', wm );
	return csize[ '.' ] * 3 + ( empty * 2 );
    }
    int p;
    int sp;
    int max;
    if( font_shadow )
    {
    }
    if( 1 )
    {
	p = cptr[ c ];
	sp = y * wm->pscreen_x + x;
	max = font->xsize * font->ysize;
	for( int yy = 0; yy < chary && p < max; yy++ )
	{
	    if( sp >= wm->pscreen_x * wm->pscreen_y ) break;
	    for( int xx = 0; xx < csize[ c ] && p < max; xx++ )
	    {
		if( font->data[ p ] > 250 ) 
		{
		    wm->g_screen[ sp ] = blend( wm->g_screen[ sp ], text_color, wm->cur_transparency );
		    if( font_bold )
		    {
			wm->g_screen[ sp + 1 ] = wm->g_screen[ sp ];
		    }
		}
		p++;
		sp++;
	    }
	    p += font->xsize - csize[ c ];
	    sp += wm->pscreen_x - csize[ c ];
	}
    }
    return csize[ c ];
}

int grayscale_img_draw_string( int x, int y, char *str, int offset, int size, image *img, window_manager *wm )
{
    int ss = 0;
    int c = offset;
    for( ; c < offset + size ; )
    {
	if( str[ c ] == 0 ) break;
	int cs = grayscale_img_draw_char( x, y, (uchar)str[ c ], img, wm );
	cs += empty;
	x += cs;
	ss += cs;
	c++;
    }
    return ss;
}

int screen_draw_string( int x, int y, char *str, int offset, int size, window_manager *wm )
{
    int ss = 0;
    int c = offset;
    for( ; c < offset + size ; )
    {
	if( str[ c ] == 0 ) break;
	int cs = screen_draw_char( x, y, (uchar)str[ c ], wm );
	if( cs ) cs += empty;
	x += cs;
	ss += cs;
	c++;
    }
    return ss;
}

void prepare_text( uchar *txt, int lsize, window_manager *wm )
{
    txt_ptr = txt;
    linesize = lsize;
    int c = 0;
    int line = 0;
    line_ptr[ line ] = 0;
    int ss = 0;
    for(;;)
    {
	if( txt[ c ] == 0 ) break;
	int cs = get_char_size( txt[ c ] );
	if( txt[ c ] == 0x0D || txt[ c ] == 0x0A ) cs = 0;
	if( cs ) cs += empty;
	ss += cs;
	if( ss > linesize || txt[ c ] == 0x0D )
	{ //out of line:
	    //search for the last SPACE:
	    for( ; c >= 0; c-- )
	    {
		if( txt[ c ] == ' ' || txt[ c ] == 0x0D || txt[ c ] == 0x0A ) break;
	    }
	    line_size[ line ] = c - line_ptr[ line ]; //size of line in a chars
	    line++;
	    line_ptr[ line ] = c + 1; //pointer to a new line
	    if( txt[ c + 1 ] == 0xD || txt[ c + 1 ] == 0xA )
		line_ptr[ line ] = c + 2;
	    ss = 0;
	}
	c++;
    }
    line_size[ line ] = c - line_ptr[ line ]; //size of the last line in a chars
    line++;
    lines = line;
}

void draw_text( image *img, window_manager *wm )
{
    /*for( int l = 0; l < lines; l++ )
    {
	if( y >= 0 && y < img->ysize - chary )
	    grayscale_img_draw_string( xoffset, y, (char*)txt_ptr, line_ptr[ l ], line_size[ l ], img, wm );
    }*/
}

void position_control( window_manager *wm )
{
    //Position control:
    int top_fade = 16;
    int bottom_fade = 16;
    int screen_top = 88;
    int screen_bottom = wm->pscreen_y - chary - bottom_fade;

    if( yoffset < 0 ) yoffset = 0;
    if( yoffset >= ( lines ) * Y_LINE_SIZE - ( screen_bottom - screen_top ) ) 
	yoffset = ( lines ) * Y_LINE_SIZE - ( screen_bottom - screen_top );
}

void screen_draw_text( int alpha, window_manager *wm )
{
    float v = (float)alpha / 1024;
    float real_speed = v * y_speed;
    yoffset += real_speed;
    if( y_speed > 0 )
    {
	y_speed -= ( v * y_fade_speed ); 
	if( y_speed < 0 ) y_speed = 0;
    }
    else
    {
	y_speed += ( v * y_fade_speed ); 
	if( y_speed > 0 ) y_speed = 0;
    }  

    int top_fade = 16;
    int bottom_fade = 16;
    int screen_top = 88;
    int screen_bottom = wm->pscreen_y - chary - bottom_fade;

    position_control( wm );

    int y = (int)( -yoffset + (float)screen_top );
    int new_page_size = 0;
    selected_line = -1;
    for( int l = 0; l < lines; l++ )
    {
	if( l & 1 ) text_color = black_text; else text_color = white_text;
	if( y >= screen_top - top_fade && y <= screen_bottom + bottom_fade )
	{
	    //Transparent fade control:
	    wm->cur_transparency = 255;
	    int mul = 256 / top_fade;
	    if( y < screen_top )
	    {
		wm->cur_transparency = ( top_fade - ( screen_top - y ) ) * mul;
	    }
	    if( y >= screen_bottom )
	    {
		wm->cur_transparency = ( bottom_fade - ( y - screen_bottom ) ) * mul;
	    }
    	    if( wm->cur_transparency > 255 ) wm->cur_transparency = 255;
	    if( wm->cur_transparency < 0 ) wm->cur_transparency = 0;

	    font_bold = 0;
	    if( font_select_y >= 0 && font_select_y >= y && font_select_y < y + Y_LINE_SIZE )
	    {
		if( txt_ptr[ line_ptr[ l ] ] != '#' )
		{
		    text_color = get_color( 130,130,130 );
		    for( int b = 0; b <= l; b++ )
		    {
			if( txt_ptr[ line_ptr[ b ] ] != '#' && 
			    txt_ptr[ line_ptr[ b ] ] != 0xD && 
			    txt_ptr[ line_ptr[ b ] ] != 0xA )
			    selected_line++;
		    }
		}
	    }
	    if( text_center )
	    {
		int ssize = get_string_size( (char*)txt_ptr, line_ptr[ l ], line_size[ l ], wm );
		screen_draw_string( ( wm->pscreen_x - ssize ) / 2, y, (char*)txt_ptr, line_ptr[ l ], line_size[ l ], wm );
	    }
	    else
		screen_draw_string( xoffset, y, (char*)txt_ptr, line_ptr[ l ], line_size[ l ], wm );
	    new_page_size++;
	}
	y += Y_LINE_SIZE;
    }
    new_page_size *= Y_LINE_SIZE;
    if( new_page_size > page_size ) page_size = new_page_size;

    //Draw page number:
    font_bold = 0;
    wm->cur_transparency = 255;
    if( !draw_page_num ) return;
    char val[ 32 ];
    char val2[ 32 ];
    int pages = lines * Y_LINE_SIZE / ( screen_bottom - screen_top );
    int_to_string( (int)yoffset / ( screen_bottom - screen_top ) + 1, val );
    int_to_string( pages, val2 );
    int xs = 0;
    mem_strcat( val, "/" );
    mem_strcat( val, val2 );
    for( int a = 0;; a++ )
    {
	if( val[ a ] == 0 ) break;
	xs += get_char_size( val[ a ] );
    }
    screen_draw_string( wm->pscreen_x - xs - 11, 7, val, 0, mem_strlen( val ), wm );

    page_size = screen_bottom - screen_top + bottom_fade;
}

void page_up( window_manager *wm )
{
    y_speed = 0;
    yoffset -= page_size;
    position_control( wm );
}

void page_down( window_manager *wm )
{
    y_speed = 0;
    yoffset += page_size;
    position_control( wm );
}
