#include "../../../sundog_engine/utils/utils.h"

int current_scene = 0;

uchar *temp_text = 0;
image *back_mask = 0;
image *back_image = 0;
image *fg_image_color = 0; //Foreground color
image *fg_image_alpha = 0; //Foreground alpha

void redraw_scene( window_manager *wm );

void create_text_buffer()
{
    temp_text = (uchar*)mem_new( HEAP_DYNAMIC, 64000, "temp text", 0 );
}

void delete_text_buffer()
{
    if( temp_text ) 
    {
	mem_free( temp_text );
	temp_text = 0;
    }
}

void load_text( char *filename )
{
    FILE *f = fopen( filename, "rb" );
    if( f )
    {
	int size = fread( temp_text, 1, 64000, f );
	temp_text[ size ] = 0;
	fclose( f );
    }
}

void draw_background( window_manager *wm )
{
    if( back_image )
    draw_image( 0, 0, wm->pscreen_x, wm->pscreen_y, back_image, wm, 0 );
}

#include <math.h>

#define MAND_TYPE float
#define MAND_PREC 0/*14*/
#define MAND_COLOR 256

struct FComplex
{
    MAND_TYPE x, y;
};

MAND_TYPE a = -0.24;
MAND_TYPE b = 0.7;

COLOR pal[ 512 ];

inline uchar GetColor( FComplex zInit )
{
    FComplex z = zInit;
    int col = 0;
    //while( ( ( z.x * z.x ) >> MAND_PREC ) + ( ( z.y * z.y ) >> MAND_PREC ) <= ( 4 << MAND_PREC ) && Color )
    for( col = 0; col < 256; col += 8 )
    {
	MAND_TYPE xx = z.x * z.x;
	MAND_TYPE yy = z.y * z.y;
	if( xx + yy > 4 ) break;
	z.y = 2 * z.x * z.y + b;
	z.x = xx - yy + a;
    }
    if( col >= 256 ) col = 255;
    return col;
}

void DrawMandelSet( MAND_TYPE xMin, 
		    MAND_TYPE xMax, 
		    MAND_TYPE yMin, 
		    MAND_TYPE yMax, 
		    window_manager *wm )
{
    MAND_TYPE xInc, yInc;
    FComplex zInit;
    int y, x;

    zInit.y = yMin;

    xInc = ( xMax - xMin ) / wm->pscreen_x;
    yInc = ( yMax - yMin ) / wm->pscreen_y;

    int ptr = 0;
    for( y = 0; y < wm->pscreen_y; y += 2, zInit.y += yInc * 2 )
    {
	zInit.x = xMin;
	for( x = 0; x < wm->pscreen_x; x += 2, zInit.x += xInc * 2 )
	{
	    COLOR col = pal[ GetColor( zInit ) + ( pseudo_random() & 7 ) ];
	    wm->g_screen[ ptr ] = col;
	    wm->g_screen[ ptr + 1 ] = col;
	    wm->g_screen[ ptr + wm->pscreen_x ] = col;
	    wm->g_screen[ ptr + wm->pscreen_x + 1 ] = col;
	    ptr += 2;
	}
	ptr += wm->pscreen_x;
    }
}

//int mas = ( 2 << MAND_PREC ) >> 1;
float mas = 0.1;

int zoom_direction = 1;

void draw_background2( int alpha, window_manager *wm )
{
    //mas -= ( 255 * alpha ) >> 10;
    alpha *= zoom_direction;
    mas += ( 0.01 * mas ) * ( (MAND_TYPE)alpha / (MAND_TYPE)1024 );
    b += ( 0.001 * b ) * ( (MAND_TYPE)alpha / (MAND_TYPE)1024 );
    a -= ( 0.001 * a ) * ( (MAND_TYPE)alpha / (MAND_TYPE)1024 );
    if( b > 0.8 ) zoom_direction = -1;
    if( b < 0.7 ) zoom_direction = 1;
    MAND_TYPE y_center = 0.4;//1;
    MAND_TYPE x_center = 0.5 + yoffset / 10000;//0.5;
    //DrawMandelSet( -2 << MAND_PREC, 1 * mas, -1 * mas, 1 * mas, wm );
    DrawMandelSet( ( -mas - y_center ) * 1.5 + y_center, ( mas - y_center ) * 1.5 + y_center, 
	           -mas + x_center, mas + x_center, wm );
}

bool first_load = false;

#define RELOAD_BACK( name ) \
    if( back_image ) clear_image( back_image ); \
    back_image = load_j2k( name, 0 );

#define RELOAD_BACK_JPG( name ) \
    if( back_image ) clear_image( back_image ); \
    back_image = load_jpeg( name, 0 );

int prev_time = 0;

void load_scene( int scene_num, window_manager *wm )
{
    y_speed = 0;
    draw_page_num = 1;
    y_dec = 1;
    black_text = get_color( 0, 0, 0 );
    white_text = get_color( 0, 0, 0 );
    int text_x_size = wm->pscreen_x - 20;
    yoffset = 0;
    text_x_size = 435;
    text_center = 0;
    current_scene = scene_num;
    if( !first_load ) 
    {
        xoffset = 10;
	first_load = true;

        fg_image_color = load_jpeg( "data/fg_color.jpg", 1 );
        fg_image_alpha = load_jpeg( "data/fg_alpha.jpg", 1 );
    }
    else
    {
        if( fg_image_color ) clear_image( fg_image_color );
        if( fg_image_alpha ) clear_image( fg_image_alpha );
	if( current_scene == 100 )
	{
	    fg_image_color = load_jpeg( "data/fg_color.jpg", 1 );
	    fg_image_alpha = load_jpeg( "data/fg_alpha.jpg", 1 );
	}
	else
	{
	    fg_image_color = load_jpeg( "data/fg_color.jpg", 1 );
	    fg_image_alpha = load_jpeg( "data/fg_alpha.jpg", 1 );
	}
    }
    switch( current_scene )
    {
	case 0:
	    RELOAD_BACK( "data/back00.j2k" );
	    load_text( "data/00.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 1:
	    RELOAD_BACK( "data/back01.j2k" );
	    load_text( "data/01.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 2:
	    RELOAD_BACK( "data/back02.j2k" );
	    load_text( "data/02.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 3:
	    RELOAD_BACK( "data/back03.j2k" );
	    load_text( "data/03.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 4:
	    black_text = get_color( 240, 240, 240 );
	    white_text = get_color( 230, 230, 230 );
	    RELOAD_BACK( "data/back04.j2k" );
	    load_text( "data/04.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 5:
	    RELOAD_BACK( "data/back05.j2k" );
	    load_text( "data/05.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 6:
	    RELOAD_BACK( "data/back06.j2k" );
	    load_text( "data/06.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 7:
	    RELOAD_BACK( "data/back07.j2k" );
	    load_text( "data/07.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 8:
	    text_x_size = 377;
	    RELOAD_BACK( "data/back08.j2k" );
	    load_text( "data/08.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 9:
	    RELOAD_BACK( "data/back09.j2k" );
	    load_text( "data/09.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 10:
	    text_x_size = 445;
	    black_text = get_color( 240, 240, 240 );
	    white_text = get_color( 230, 230, 230 );
	    RELOAD_BACK( "data/back10.j2k" );
	    load_text( "data/10.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 11:
	    RELOAD_BACK( "data/back11.j2k" );
	    load_text( "data/11.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 12:
	    RELOAD_BACK( "data/back12.j2k" );
	    load_text( "data/12.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 13:
	    text_x_size = 377;
	    RELOAD_BACK( "data/back13.j2k" );
	    load_text( "data/13.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 14:
	    text_x_size = 425;
	    RELOAD_BACK( "data/back14.j2k" );
	    load_text( "data/14.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 15:
	    text_x_size = 377;
	    RELOAD_BACK( "data/back15.j2k" );
	    load_text( "data/15.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 16:
	    RELOAD_BACK( "data/back16.j2k" );
	    load_text( "data/16.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 17:
	    text_x_size = 425;
	    RELOAD_BACK_JPG( "data/back17.jpg" );
	    load_text( "data/17.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 18:
	    text_x_size = 377;
	    black_text = get_color( 240, 240, 240 );
	    white_text = get_color( 230, 230, 230 );
	    RELOAD_BACK( "data/back18.j2k" );
	    load_text( "data/18.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 19:
	    RELOAD_BACK( "data/back19.j2k" );
	    load_text( "data/19.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 20:
	    text_x_size = 377;
	    RELOAD_BACK( "data/back20.j2k" );
	    load_text( "data/20.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 21:
	    RELOAD_BACK( "data/back21.j2k" );
	    load_text( "data/21.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 22:
	    text_x_size = 377;
	    RELOAD_BACK( "data/back22.j2k" );
	    load_text( "data/22.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 23:
	    RELOAD_BACK( "data/back23.j2k" );
	    load_text( "data/23.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 24:
	    text_x_size = 370;
	    RELOAD_BACK( "data/back24.j2k" );
	    load_text( "data/24.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 25:
	    RELOAD_BACK( "data/back25.j2k" );
	    load_text( "data/25.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 26:
	    RELOAD_BACK( "data/back26.j2k" );
	    load_text( "data/26.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 27:
	    RELOAD_BACK( "data/back27.j2k" );
	    load_text( "data/27.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 28:
	    RELOAD_BACK( "data/back28.j2k" );
	    load_text( "data/28.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 29:
	    RELOAD_BACK( "data/back29.j2k" );
	    load_text( "data/29.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 30:
	    RELOAD_BACK( "data/back30.j2k" );
	    load_text( "data/30.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    redraw_scene( wm );
	    break;
	case 100:
	    for( int pp = 0, ppp; pp < 512; pp++ )
	    {
		ppp = pp;
		int r = ppp;
		int g = ppp + 100;
		int b = ppp + 190;
		if( r > 255 ) r = 255;
		if( g > 255 ) g = 255;
		if( b > 255 ) b = 255;
		pal[ pp ] = blend( get_color( r, g, b ), COLORMASK, 64 );
	    }
	    y_dec = 3;
	    draw_page_num = 0;
	    font_shadow = 0;
	    text_center = 1;
	    if( back_image ) { clear_image( back_image ); back_image = 0; }
	    load_text( "data/txt_.txt" );
	    prepare_text( temp_text, text_x_size, wm );
	    prev_time = time_ticks();
	    redraw_scene( wm );
	    break;
    }
}

int fps = 0;
int fps_cnt = 0;
int fps_time = 0;

void redraw_scene( window_manager *wm )
{
    fps_cnt++;
    if( fps_time != time_seconds() )
    {
	fps_time = time_seconds();
	fps = fps_cnt;
	fps_cnt = 0;
    }
    char fps_str[ 64 ];
    int_to_string( fps, fps_str );

    int cur_time = time_ticks();
    int alpha = cur_time - prev_time;
    alpha <<= 10;
    alpha /= time_ticks_per_second();
    //alpha: 1024 - one second
    prev_time = cur_time;
    switch( current_scene )
    {
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 13:
	case 14:
	case 15:
	case 16:
	case 17:
	case 18:
	case 19:
	case 20:
	case 21:
	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 27:
	case 28:
	case 29:
	case 30:
	    draw_background( wm );
	    screen_draw_text( alpha, wm );
	    grayscale_draw_image_with_alpha( 0, 0, fg_image_color, fg_image_alpha, wm );
	    break;
	case 100:
	    draw_background2( alpha, wm );
	    screen_draw_text( alpha, wm );
	    grayscale_draw_image_with_alpha( 0, 0, fg_image_color, fg_image_alpha, wm );
	    break;
    }
    //draw_string( fps_str, 1, 1, wm );
}
