/*
    SunDog: user_code.cpp.
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***                      observer_page@mail.ru
//***               WWW: warmplace.ru

#include "../../../sundog_engine/main/user_code.h"
#include "../../../sundog_engine/filesystem/v3nus_fs.h"
#include "../../../sundog_engine/time/timemanager.h"

#include "destiny_jpg.h"
#include "destiny_font.h"
#include "destiny_music.h"
#include "destiny_scenes.h"

#ifdef LINUX
    char *user_window_name = "Otzvuki Sudby #4 (2007)";
#else
    char *user_window_name = "  #4 (2007)";
#endif
char *user_config_file_name = "config.ini";
int user_window_xsize = 640;
int user_window_ysize = 480;
int user_window_flags = 0;
void *user_sound_data = 0;

int desktop2_handler( sundog_event *evt, window_manager *wm )
{
    WINDOWPTR win = evt->event_win;
    int retval = 0;
    switch( evt->event_type )
    {
	case EVT_DRAW: retval = 1; break;
	case EVT_SCREENRESIZE: 
	    win->xsize = wm->screen_xsize;
	    win->ysize = wm->screen_ysize;
	    direct_draw_resize( win, win->xsize, win->ysize, wm );
	    retval = 1;
	    break;
    }
    return retval;
}

void user_init( window_manager *wm )
{
    //Windows Direct Draw init:
    new_window( "desktop", 0, 0, wm->screen_xsize, wm->screen_ysize, get_color(255,0,0), 0, desktop2_handler, wm );
    show_window( wm->root_win, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    direct_draw_init( wm->root_win, wm->screen_xsize, wm->screen_ysize, wm );
    direct_draw_set_current_window( wm->root_win, wm );

    //Start sound stream:
    music_init();

    //User init:
    load_font( "data/font_288.jpg", 32, 8 );
    //load_font( "data/font_211.jpg", 12, 22 );
    /*FILE *f = fopen( "chars.txt", "wb" );
    if( f )
    {
	int r, g;
	int cc = 0;
	for( g = 0; g < 8; g++ )
	{
	    for( r = 0; r < 32; r++ ) fprintf( f, "%c", cc++ );
	    fprintf( f, "\n" );
	}
	fclose( f );
    }*/

    create_text_buffer();
    load_scene( 100, wm );
    draw_screen( wm );
}

float const_speed = 0;
float menu_yoffset = 0;

int drag_y_start = 0;
float start_offset = 0;

int user_event_handler( sundog_event *evt, window_manager *wm )
{
    if( evt->event_type == EVT_BUTTONUP )
    {
	const_speed = 0;
    }
    if( evt->event_type == EVT_BUTTONDOWN )
    {
	if( ( evt->button & 511 ) == KEY_PAGEDOWN ) 
	{ 
	    if( evt->button & KEY_SHIFT )
		page_down( wm );
	    else
		const_speed = 200;
	}
	if( ( evt->button & 511 ) == KEY_PAGEUP ) 
	{ 
	    if( evt->button & KEY_SHIFT )
		page_up( wm );
	    else
		const_speed = -200;
	}
	if( ( evt->button & 511 ) == KEY_DOWN ) { const_speed = 80; }
	if( ( evt->button & 511 ) == KEY_UP ) { const_speed = -80; }
	if( ( evt->button & 511 ) == KEY_ESCAPE ) 
	{
	    if( current_scene < 100 )
	    {
		load_scene( 100, wm );
		yoffset = menu_yoffset;
	    }
	    else wm->exit_request = 1;
	}
	switch( evt->button )
	{
	    case '1': change_track( 0 ); break;
	    case '2': change_track( 1 ); break;
	    case '3': change_track( 2 ); break;
	    case '4': change_track( 3 ); break;
	    case '5': change_track( 4 ); break;
	    case '6': change_track( 5 ); break;
	    case '7': change_track( 6 ); break;
	    case '8': change_track( 7 ); break;
	    case '9': change_track( 8 ); break;
	    case '0': change_track( 255 ); break;
	}
    }
    if( evt->event_type == EVT_MOUSEMOVE )
    {
	if( evt->mouse_button == BUTTON_LEFT )
	{
	    if( current_scene == 100 )
	    {
		font_select_y = evt->y;
	    }
	    else
	    {
		int d = evt->y - drag_y_start;
		yoffset = start_offset - d;
		const_speed = 0;
		y_speed = 0;
	    }
	}
    }
    if( evt->event_type == EVT_MOUSEBUTTONDOWN )
    {
	if( evt->mouse_button == BUTTON_SCROLLDOWN ) 
	{ 
	    page_down( wm );
	}
	if( evt->mouse_button == BUTTON_SCROLLUP ) 
	{ 
	    page_up( wm );
	}
	if( evt->mouse_button == BUTTON_LEFT )
	{
	    if( current_scene == 100 )
	    {
		font_select_y = evt->y;
	    }
	    else
	    {
		drag_y_start = evt->y;
		start_offset = yoffset;
	    }
	    redraw_scene( wm );
	}
    }
    if( evt->event_type == EVT_MOUSEBUTTONUP )
    {
	if( evt->mouse_button == BUTTON_RIGHT )
	{
	    if( current_scene < 100 )
	    {
		load_scene( 100, wm );
		yoffset = menu_yoffset;
	    }
	}
	if( evt->mouse_button == BUTTON_LEFT )
	{
	    if( current_scene == 100 )
	    {
		//Make some action:
		if( selected_line >= 0 )
		{
		    font_select_y = -1;
		    menu_yoffset = yoffset;
		    load_scene( selected_line, wm );
		}
	    }
	}
    }
    return 1;
}

void user_screen_redraw( window_manager *wm )
{
    if( const_speed ) y_speed = const_speed;
    music_control();
    redraw_scene( wm );
    draw_screen( wm );
}

void user_close( window_manager *wm )
{
    music_close();
}
