/*
    SunDog: main.cpp. SunDog Engine main()
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***               WWW: warmplace.ru

#ifdef NONPALM
    #include <stdio.h>
#endif

#include "../user_code.h"

//################################
//## DEVICE VARIABLES:          ##
//################################

//PALMOS
#ifndef NONPALM
    #include <PalmOS.h>
#ifndef NATIVEARM
    SysAppInfoPtr SysGetAppInfo(SysAppInfoPtr *rootAppPP,
	                        SysAppInfoPtr *actionCodeAppPP)
    	                        SYS_TRAP(sysTrapSysUIBusy);
#else
    #define arm_startup __attribute__ ((section ("arm_startup")))
    #include "palm_functions.h"
#endif //NATIVEARM
#endif

//################################
//################################
//################################

//################################
//## APPLICATION MAIN:          ##
//################################

extern char *user_window_name;
extern char *user_config_file_name;
extern int user_window_xsize;
extern int user_window_ysize;
extern int user_window_flags;

window_manager wm;
int g_frame = 0;
int g_skip_frames = 0;

//********************************
//WIN32 MAIN *********************
#ifdef WIN
int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPSTR lpszCmdLine, int nCmdShow)
{
    {
	wm.hCurrentInst = hCurrentInst;
	wm.hPreviousInst = hPreviousInst; 
	wm.lpszCmdLine = lpszCmdLine;
	wm.nCmdShow = nCmdShow;
#endif
//********************************
//********************************

//********************************
//PALMOS MAIN ********************
#ifndef NONPALM
#ifndef NATIVEARM
UInt32 PilotMain(UInt16 launchCode, void *cmdPBP, UInt16 launchFlags)
{
    if( launchCode == sysAppLaunchCmdSyncNotify )
    {
    }
    if( launchCode == sysAppLaunchCmdNormalLaunch )
    {
	appInfo = SysGetAppInfo( &ai1, &ai2 );
	ownID = appInfo->memOwnerID;
#else
long ARM_PalmOS_main( const void *emulStateP, void *userData, Call68KFuncType *call68KFuncP ) arm_startup;
long ARM_PalmOS_main( const void *emulStateP, void *userData, Call68KFuncType *call68KFuncP )
{
    {
	void *oldGOT;
	register void *gGOT asm ("r10");
	ARM_INFO *arm_info = (ARM_INFO *)userData;
	oldGOT = (void*)gGOT;
	gGOT = (void *)arm_info->GOT;
	ownID = (unsigned short)arm_info->ID;
	form_handler = arm_info->FORMHANDLER; //form_handler defined in palm_functions.cpp
	CALL_INIT
	dprint( "MAIN: ARM started\n" );
#endif //NATIVEARM
	int autooff_time = SysSetAutoOffTime( 0 );
#endif //NONPALM
//********************************
//********************************

//********************************
//LINUX MAIN *********************
#ifdef LINUX
int main(int argc, char *argv[])
{
    {
	wm.argc = argc;
	wm.argv = argv;
	g_skip_frames = 1;
#ifdef TEXTMODE
	char* str;
	if( argc > 1 )
	{
	    str = argv[ 1 ];
	    if( str[ 0 ] == '-' || str[ 0 ] == '?' )
	    {
		printf( "\033[%dm\033[%dm", 37, 40 );
		printf( "PsyTexx 2 for Linux.\n" );
		printf( "\033[%dm\033[%dm", 36, 40 );
		printf( "Command line options:\n" );
		printf( "\033[%dm\033[%dm", 35, 40 );
		printf( "m - auto mouse key up (for some GPM versions)\n" );
		printf( "\033[%dm\033[%dm", 37, 40 );
		return 0;
	    }
	    if( str[ 0 ] == 'm' ) mouse_auto_keyup = 1;
	}
#endif
#endif
//********************************
//********************************

	debug_reset();

	dprint( "\n" );
	dprint( "\n" );
	dprint( "%s\n", PROG_VERSION );
	dprint( "%s\n", PROG_DATE );
	dprint( "\n" );
	dprint( "STARTING...\n" );
	dprint( "\n" );
	dprint( "\n" );

	get_disks();
	read_file_with_options( user_config_file_name );
	win_init( user_window_name, user_window_xsize, user_window_ysize, user_window_flags, &wm );
#ifndef NOSOUND
	sound_stream_init( 44100, 2 );
	sound_stream_play();
#endif
	user_init( &wm );
	while( 1 )
	{
	    window_manager *wmp = &wm;
	    FIRST_PART_OF_THE_EVENT_LOOP( wmp )
	    }
	    LAST_PART_OF_THE_EVENT_LOOP( wmp )
	}
        user_close( &wm );
#ifndef NOSOUND
	sound_stream_close(); //Close sound stream
#endif
	win_close( &wm );     //Close window manager
	debug_close();
	mem_free_all();       //Close all memory blocks

	dprint( "\n" );
	dprint( "\n" );
	dprint( "BYE !\n" );
	dprint( "\n" );
	dprint( "\n" );

#ifndef NONPALM
	SysSetAutoOffTime( autooff_time );
#endif
#ifdef NATIVEARM
	gGOT = (void*)oldGOT;
#endif
    }

    return 0;
}

//################################
//################################
//################################
