/*
    SunDog: utils.cpp. Different things: string list; random generator
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***                      observer_page@mail.ru
//***               WWW: warmplace.ru

#include "../../core/core.h"
#include "../../memory/memory.h"
#include "../utils.h"

//WORKING WITH A STRING LISTS:

void list_init( list_data *data )
{
    data->items = (char*)mem_new( HEAP_STORAGE, MAX_SIZE, "list items", 0 );
    data->items_ptr = (long*)mem_new( HEAP_STORAGE, MAX_ITEMS * 4, "list items ptrs", 0 );
    list_clear( data );
    list_reset_selection( data );
}

void list_close( list_data *data )
{
    if( data->items )
    {
	mem_free( data->items );
	data->items = 0;
    }
    if( data->items_ptr )
    {
	mem_free( data->items_ptr );
	data->items_ptr = 0;
    }
}

void list_clear( list_data *data )
{
    data->items_num = 0;
    data->start_item = 0;
}

void list_reset_selection( list_data *data )
{
    data->selected_item = -1;
}

void list_add_item( char *item, char attr, list_data *data )
{
    long ptr, p;

    mem_off();
    
    if( data->items_num == 0 ) 
    {
	data->items_ptr[ 0 ] = 0;
	ptr = 0;
    }
    else 
    {
	ptr = data->items_ptr[ data->items_num ];
    }
    
    data->items[ ptr++ ] = attr | 128;
    for( p = 0; ; p++, ptr++ )
    {
	data->items[ ptr ] = item[ p ];
	if( item[ p ] == 0 ) break;
    }
    ptr++;
    data->items_num++;
    data->items_ptr[ data->items_num ] = ptr;
    
    mem_on();
}

void list_delete_item( long item_num, list_data *data )
{
    long ptr, p;

    if( item_num < data->items_num && item_num >= 0 )
    {
	mem_off();
	
	ptr = data->items_ptr[ item_num ]; //Offset of our item (in the "items")
	
	//Get item size (in chars):
	long size;
	for( size = 0; ; size++ )
	{
	    if( data->items[ ptr + size ] == 0 ) break;
	}
	
	//Delete it:
	for( p = ptr; p < MAX_SIZE - size - 1 ; p++ )
	{
	    data->items[ p ] = data->items[ p + size + 1 ];
	}
	for( p = 0; p < data->items_num; p++ )
	{
	    if( data->items_ptr[ p ] > ptr ) data->items_ptr[ p ] -= size + 1;
	}
	for( p = item_num; p < data->items_num; p++ )
	{
	    if( p + 1 < MAX_ITEMS ) 
		data->items_ptr[ p ] = data->items_ptr[ p + 1 ];
	    else
		data->items_ptr[ p ] = 0;
	}
	data->items_num--;
	if( data->items_num < 0 ) data->items_num = 0;

	if( data->selected_item >= data->items_num ) data->selected_item = data->items_num - 1;
	
	mem_on();
    }
}

void list_move_item_up( long item_num, list_data *data )
{
    if( item_num < data->items_num && item_num >= 0 )
    {
	mem_off();
	if( item_num != 0 )
	{
	    long temp = data->items_ptr[ item_num - 1 ];
	    data->items_ptr[ item_num - 1 ] = data->items_ptr[ item_num ];
	    data->items_ptr[ item_num ] = temp;
	    if( item_num == data->selected_item ) data->selected_item--;
	}
	mem_on();
    }
}	

void list_move_item_down( long item_num, list_data *data )
{
    if( item_num < data->items_num && item_num >= 0 )
    {
	mem_off();
	if( item_num != data->items_num - 1 )
	{
	    long temp = data->items_ptr[ item_num + 1 ];
	    data->items_ptr[ item_num + 1 ] = data->items_ptr[ item_num ];
	    data->items_ptr[ item_num ] = temp;
	    if( item_num == data->selected_item ) data->selected_item++;
	}
	mem_on();
    }
}

char* list_get_item( long item_num, list_data *data )
{
    if( item_num >= data->items_num ) return 0;
    if( item_num >= 0 )
	return data->items + data->items_ptr[ item_num ] + 1;
    else 
	return 0;
}

char list_get_attr( long item_num, list_data *data )
{
    if( item_num >= data->items_num ) return 0;
    if( item_num >= 0 )
	return data->items[ data->items_ptr[ item_num ] ] & 127;
    else
	return 0;
}

long list_get_selected_num( list_data *data )
{
    return data->selected_item;
}

void list_set_selected_num( long sel, list_data *data )
{
    data->selected_item = sel;
}

//Return values:
//1 - item1 > item2
//0 - item1 <= item2
long list_compare_items( long item1, long item2, list_data *data )
{
    char *i1 = data->items + data->items_ptr[ item1 ];
    char *i2 = data->items + data->items_ptr[ item2 ];
    char a1 = i1[ 0 ] & 127;
    char a2 = i2[ 0 ] & 127;
    i1++;
    i2++;
    
    long retval = 0;
    
    //Compare:
    if( a1 != a2 )
    {
	if( a1 == 1 ) retval = 0;
	if( a2 == 1 ) retval = 1;
    }
    else
    {
	for( int a = 0; ; a++ )
	{
	    if( i1[ a ] == 0 ) break;
	    if( i2[ a ] == 0 ) break;
	    if( i1[ a ] < i2[ a ] ) { break; }             //item1 < item2
	    if( i1[ a ] > i2[ a ] ) { retval = 1; break; } //item1 > item2
	}
    }
    
    return retval;
}

void list_sort( list_data *data )
{
    mem_off();
    for(;;)
    {
	int s = 0;
	for( long a = 0; a < data->items_num - 1; a++ )
	{
	    if( list_compare_items( a, a + 1, data ) )
	    {
		s = 1;
		long temp = data->items_ptr[ a + 1 ];
		data->items_ptr[ a + 1 ] = data->items_ptr[ a ];
		data->items_ptr[ a ] = temp;
	    }
	}
	if( s == 0 ) break;
    }
    mem_on();
}

//RANDOM GENERATOR:

unsigned long rand_next = 1;

void set_seed( unsigned long seed )
{
    rand_next = seed;
}

unsigned long pseudo_random( void )
{
    rand_next = rand_next * 1103515245 + 12345;
    return ((unsigned int)(rand_next/65536)%32768);
}

//WORKING WITH A STRINGS:

void int_to_string( int value, char *str )
{
    int n, p = 0, f = 0;
    
    if( value < 0 )
    {
    	str[ p++ ] = '-';
	value = -value;
    }
    
    n = value / 10000000; if( n || f ) { str[p++] = (char)n+48; value -= n * 10000000; f = 1; }
    n = value / 1000000; if( n || f ) { str[p++] = (char)n+48; value -= n * 1000000; f = 1; }
    n = value / 100000; if( n || f ) { str[p++] = (char)n+48; value -= n * 100000; f = 1; }
    n = value / 10000; if( n || f ) { str[p++] = (char)n+48; value -= n * 10000; f = 1; }
    n = value / 1000; if( n || f ) { str[p++] = (char)n+48; value -= n * 1000; f = 1; }
    n = value / 100; if( n || f ) { str[p++] = (char)n+48; value -= n * 100; f = 1; }
    n = value / 10; if( n || f ) { str[p++] = (char)n+48; value -= n * 10; f = 1; }
    str[ p++ ] = (char)value + 48;
    str[ p ] = 0;
}

int string_to_int( char *str )
{
    int res = 0;
    int d = 1;
    int minus = 1;
    for( int a = mem_strlen( str ) - 1; a >= 0; a-- )
    {
	int v = str[ a ];
	if( v >= '0' && v <= '9' )
	{
	    v -= '0';
	    res += v * d;
	    d *= 10;
	}
	else
	if( v == '-' ) minus = -1;
    }
    return res * minus;
}

int hex_string_to_int( char *str )
{
    int res = 0;
    int d = 1;
    int minus = 1;
    for( int a = mem_strlen( str ) - 1; a >= 0; a-- )
    {
	int v = str[ a ];
	if( ( v >= '0' && v <= '9' ) || ( v >= 'A' && v <= 'F' ) || ( v >= 'a' && v <= 'f' ) )
	{
	    if( v >= '0' && v <= '9' ) v -= '0';
	    else
	    if( v >= 'A' && v <= 'F' ) { v -= 'A'; v += 10; }
	    else
	    if( v >= 'a' && v <= 'f' ) { v -= 'a'; v += 10; }
	    res += v * d;
	    d *= 16;
	}
	else
	if( v == '-' ) minus = -1;
    }
    return res * minus;
}
