/*
    SunDog: directdraw.cpp. Functions for direct drawing in buffer of some window
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***               WWW: warmplace.ru

#include "../../main/user_code.h"
#include "../../core/debug.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

extern unsigned char font8x8[ 2050 ];

void draw_screen( window_manager *wm )
{
    direct_draw( wm->cur_window, wm );
}

void clear_screen( window_manager *wm )
{
    if( wm->cur_window && wm->cur_window->ddraw_buf )
	mem_set( wm->cur_window->ddraw_buf, wm->cur_window->ddraw_xsize * wm->cur_window->ddraw_ysize * COLORLEN, 0 );
}

#define bottom 1
#define top 2
#define left 4
#define right 8
int make_code( int x, int y, int clip_x, int clip_y )
{
    int code = 0;
    if( y >= clip_y ) code = bottom;
    else if( y < 0 ) code = top;
    if( x >= clip_x ) code += right;
    else if( x < 0 ) code += left;
    return code;
}

void draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm )
{
    //Cohen-Sutherland line clipping algorithm:
    int code0;
    int code1;
    int out_code;
    int x, y;
    code0 = make_code( x1, y1, wm->pscreen_x, wm->pscreen_y );
    code1 = make_code( x2, y2, wm->pscreen_x, wm->pscreen_y );
    while( code0 || code1 )
    {
	if( code0 & code1 ) return; //Trivial reject
	else
	{
	    //Failed both tests, so calculate the line segment to clip
	    if( code0 )
		out_code = code0; //Clip the first point
	    else
		out_code = code1; //Clip the last point

	    if( out_code & bottom )
	    {
		//Clip the line to the bottom of the viewport
		y = wm->pscreen_y - 1;
		x = x1 + ( x2 - x1 ) * ( y - y1 ) / ( y2 - y1 );
	    }
	    else 
	    if( out_code & top )
	    {
		y = 0;
		x = x1 + ( x2 - x1 ) * ( y - y1 ) / ( y2 - y1 );
	    }
	    else
	    if( out_code & right )
	    {
		x = wm->pscreen_x - 1;
		y = y1 + ( y2 - y1 ) * ( x - x1 ) / ( x2 - x1 );
	    }
	    else
	    if( out_code & left )
	    {
		x = 0;
		y = y1 + ( y2 - y1 ) * ( x - x1 ) / ( x2 - x1 );
	    }

	    if( out_code == code0 )
	    { //Modify the first coordinate 
		x1 = x; y1 = y;
		code0 = make_code( x1, y1, wm->pscreen_x, wm->pscreen_y );
	    }
	    else
	    { //Modify the second coordinate
		x2 = x; y2 = y;
		code1 = make_code( x2, y2, wm->pscreen_x, wm->pscreen_y );
	    }
	}
    }

    //Draw line:
    int len_x = x2 - x1; if( len_x < 0 ) len_x = -len_x;
    int len_y = y2 - y1; if( len_y < 0 ) len_y = -len_y;
    int ptr = y1 * wm->pscreen_x + x1;
    int delta;
    int v = 0, old_v = 0;
    int transp = wm->cur_transparency;
    if( len_x > len_y )
    {
	//Horisontal:
	if( len_x != 0 )
	    delta = ( len_y << 10 ) / len_x;
	else
	    delta = 0;
	if( transp >= 255 )
	    for( int a = 0; a <= len_x; a++ )
	    {
		wm->g_screen[ ptr ] = color;
		old_v = v;
		v += delta;
		if( x2 - x1 > 0 ) ptr++; else ptr--;
		if( ( old_v >> 10 ) != ( v >> 10 ) ) 
		{
		    if( y2 - y1 > 0 )
			ptr += wm->pscreen_x;
		    else
			ptr -= wm->pscreen_x;
		}
	    }
	else
	    for( int a = 0; a <= len_x; a++ )
	    {
		wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], color, transp );
		old_v = v;
		v += delta;
		if( x2 - x1 > 0 ) ptr++; else ptr--;
		if( ( old_v >> 10 ) != ( v >> 10 ) ) 
		{
		    if( y2 - y1 > 0 )
			ptr += wm->pscreen_x;
		    else
			ptr -= wm->pscreen_x;
		}
	    }
    }
    else
    {
	//Vertical:
	if( len_y != 0 ) 
	    delta = ( len_x << 10 ) / len_y;
	else
	    delta = 0;
	if( transp >= 255 )
	    for( int a = 0; a <= len_y; a++ )
	    {
		wm->g_screen[ ptr ] = color;
		old_v = v;
		v += delta;
		if( y2 - y1 > 0 ) 
		    ptr += wm->pscreen_x;
		else
		    ptr -= wm->pscreen_x;
		if( ( old_v >> 10 ) != ( v >> 10 ) ) 
		{
		    if( x2 - x1 > 0 ) ptr++; else ptr--;
		}
	    }
	else
	    for( int a = 0; a <= len_y; a++ )
	    {
		wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], color, transp );
		old_v = v;
		v += delta;
		if( y2 - y1 > 0 ) 
		    ptr += wm->pscreen_x;
		else
		    ptr -= wm->pscreen_x;
		if( ( old_v >> 10 ) != ( v >> 10 ) ) 
		{
		    if( x2 - x1 > 0 ) ptr++; else ptr--;
		}
	    }
    }
}

void draw_frame( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm )
{
    draw_line( x, y, x + x_size, y, color, wm );
    draw_line( x + x_size, y, x + x_size, y + y_size, color, wm );
    draw_line( x + x_size, y + y_size, x, y + y_size, color, wm );
    draw_line( x, y + y_size, x, y, color, wm );
}

void draw_box( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm )
{
    if( x < 0 ) { x_size += x; x = 0; }
    if( y < 0 ) { y_size += y; y = 0; }
    if( x + x_size <= 0 ) return;
    if( y + y_size <= 0 ) return;
    if( x + x_size > wm->pscreen_x ) x_size -= x + x_size - wm->pscreen_x;
    if( y + y_size > wm->pscreen_y ) y_size -= y + y_size - wm->pscreen_y;
    if( x >= wm->pscreen_x ) return;
    if( y >= wm->pscreen_y ) return;
    if( x_size < 0 ) return;
    if( y_size < 0 ) return;

    int ptr = y * wm->pscreen_x + x;
    int transp = wm->cur_transparency;
    if( transp >= 255 )
	for( int cy = 0; cy < y_size; cy++ )
	{
	    for( int cx = 0; cx < x_size; cx++ )
	    {
		wm->g_screen[ ptr ] = color;
		ptr++;
	    }
	    ptr += wm->pscreen_x - x_size;
	}
    else
	for( int cy = 0; cy < y_size; cy++ )
	{
	    for( int cx = 0; cx < x_size; cx++ )
	    {
		wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], color, transp );
		ptr++;
	    }
	    ptr += wm->pscreen_x - x_size;
	}
}

int draw_char( uchar c, int x, int y, window_manager *wm )
{
    //Bounds control:
    if( c == 0 ) return char_x_size( c, wm );
    if( x < 0 ) return char_x_size( c, wm );
    if( y < 0 ) return char_x_size( c, wm );
    if( x > wm->pscreen_x - char_x_size( c, wm ) ) return char_x_size( c, wm );
    if( y > wm->pscreen_y - char_y_size( wm ) ) return char_x_size( c, wm );

    int sp = y * wm->pscreen_x + x;

    int p = c * 8;

    int transp = wm->cur_transparency;

    uchar invert_mask = 0;
    if( transp >= 255 )
    {
	if( wm->cur_font_invert ) invert_mask = 0xFF;
	for( int cy = 0; cy < 8; cy++ )
	{
	    int fpart = font8x8[ p ] ^ invert_mask;
	    for( int cx = 0; cx < 8; cx++ )
	    {
		if( fpart & 128 ) wm->g_screen[ sp ] = wm->cur_font_color;
		    else if( wm->cur_font_draw_bgcolor ) wm->g_screen[ sp ] = wm->cur_font_bgcolor;
		fpart <<= 1;
		sp++;
	    }
	    sp += wm->pscreen_x - 8;
	    p++;
	}
    }
    else
    {
	if( wm->cur_font_invert ) invert_mask = 0xFF;
	for( int cy = 0; cy < 8; cy++ )
	{
	    int fpart = font8x8[ p ] ^ invert_mask;
	    for( int cx = 0; cx < 8; cx++ )
	    {
		if( fpart & 128 ) wm->g_screen[ sp ] = blend( wm->g_screen[ sp ], wm->cur_font_color, transp );
		else if( wm->cur_font_draw_bgcolor ) wm->g_screen[ sp ] = blend( wm->g_screen[ sp ], wm->cur_font_bgcolor, transp );
		fpart <<= 1;
		sp++;
	    }
	    sp += wm->pscreen_x - 8;
	    p++;
	}
    }
    return char_x_size( c, wm );
}

void draw_string( char *str, int x, int y, window_manager *wm )
{
    //Bounds control:
    if( y < 0 ) return;
    if( x > wm->pscreen_x ) return;

    int start_x = x;

    int p = 0;
    while( str[ p ] )
    {
	if( str[ p ] == 0xA ) 
	{ 
	    y += char_y_size( wm ); x = start_x; 
	}
	else
	{
	    if( str[ p ] != 0xD ) 
		draw_char( (unsigned char)str[ p ], x, y, wm );
	    x += char_x_size( (unsigned char)str[ p ], wm );
	}
	p++;
    }
}

