/*
    SunDog: handlers.cpp. Standart window handlers
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***               WWW: warmplace.ru

#include "../../core/debug.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../../main/user_code.h"
#include "../wmanager.h"

int desktop_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    switch( evt->event_type )
    {
	case EVT_SCREENRESIZE:
	    win->x = 0;
	    win->y = 0;
	    win->xsize = wm->screen_xsize;
	    win->ysize = wm->screen_ysize;
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	    retval = 1;
	    break;
    }
    return retval;
}

int label_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    switch( evt->event_type )
    {
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	    if( wm->prev_focus_win && wm->prev_focus_win->visible )
		wm->focus_win = wm->prev_focus_win;
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->parent->color, wm );
	    //Draw name:
	    {
		int ty = ( win->ysize - char_y_size( wm ) ) / 2;
		win_draw_string( win, win->name, 0, ty, get_color(0,0,0), win->parent->color, wm );
	    }
	    retval = 1;
	    break;
    }
    return retval;
}

struct text_data 
{
    uchar *text;
    int cur_pos;
    int active;
};

int text_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    text_data *data = (text_data*)win->data;
    COLOR col, col2;
    int p, s, b, i;
    int xp;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( text_data );
	    break;
	case EVT_AFTERCREATE:
	    data->text = (uchar*)MEM_NEW( HEAP_DYNAMIC, 32 );
	    data->text[ 0 ] = 0;
	    data->cur_pos = 0;
	    data->active = 0;
	    break;
	case EVT_UNFOCUS:
	    data->active = 0;
	    draw_window( win, wm );
	    retval = 1;
	    break;
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONDOWN:
	    if( evt->mouse_button & BUTTON_LEFT )
	    {
		data->active = 1;
		xp = 1;
		for( p = 0; p < mem_get_size( data->text ); p++ )
		{
		    s = data->text[ p ];
		    int csize = char_x_size( s, wm );
		    if( s == 0 ) csize = 300;
		    if( rx >= xp && rx < xp + csize )
		    {
			data->cur_pos = p;
			break;
		    }
		    xp += char_x_size( s, wm );
		    if( data->text[ p ] == 0 ) break;
		}
		draw_window( win, wm );
	    }
	case EVT_MOUSEBUTTONUP:
	    if( data->active ) 
		retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( data->active == 0 ) 
	    {
		retval = 0;
		break;
	    }
	    b = evt->button;
	    if( b == KEY_ENTER || b == KEY_ESCAPE )
	    {
		if( b == KEY_ENTER && win->action_handler )
		    win->action_handler( win->handler_data, win, wm );
		data->active = 0;
		if( wm->prev_focus_win && wm->prev_focus_win->visible )
		    wm->focus_win = wm->prev_focus_win;
		draw_window( win, wm );
		retval = 1;
		break;
	    }
	    if( b & KEY_SHIFT )
	    {
		b &= 255;
		if( b >= 'a' && b <= 'z' ) b -= 0x20;
		else
		{
		    switch( b )
		    {
			case '0': b = ')'; break;
			case '1': b = '!'; break;
			case '2': b = '@'; break;
			case '3': b = '#'; break;
			case '4': b = '$'; break;
			case '5': b = '%'; break;
			case '6': b = '^'; break;
			case '7': b = '&'; break;
			case '8': b = '*'; break;
			case '9': b = '('; break;
			case '[': b = '{'; break;
			case ']': b = '}'; break;
			case ';': b = ':'; break;
			case  39: b = '"'; break;
			case ',': b = '<'; break;
			case '.': b = '>'; break;
			case '/': b = '?'; break;
			case '-': b = '_'; break;
			case '=': b = '+'; break;
			case  92: b = '|'; break;
			case '`': b = '~'; break;
		    }
		}
	    }	    
	    if( b == KEY_BACKSPACE )
	    {
		if( data->cur_pos >= 1 )
		{
		    data->cur_pos--;
		    for( i = data->cur_pos; i < mem_get_size( data->text ); i++ ) data->text[ i ] = data->text[ i + 1 ];
		}
	    }
	    else
	    if( b == KEY_DELETE )
	    {
		if( data->text[ data->cur_pos ] != 0 )
		for( i = data->cur_pos; i < mem_get_size( data->text ); i++ ) data->text[ i ] = data->text[ i + 1 ];
	    }
	    else
	    if( b == KEY_LEFT )
	    {
		if( data->cur_pos >= 1 )
		{
		    data->cur_pos--;
		}
	    }
	    else
	    if( b == KEY_RIGHT )
	    {
		if( data->text[ data->cur_pos ] != 0 )
		    data->cur_pos++;
	    }
	    else
	    if( b == KEY_END )
	    {
		for( i = 0; i < mem_get_size( data->text ); i++ ) if( data->text[ i ] == 0 ) break;
		data->cur_pos = i;
	    }
	    else
	    if( b == KEY_HOME )
	    {
		data->cur_pos = 0;
	    }
	    else
	    {
		for( i = mem_get_size( data->text ) - 1; i >= data->cur_pos + 1; i-- )
		    data->text[ i ] = data->text[ i - 1 ];
		data->text[ data->cur_pos ] = (uchar)b;
		data->cur_pos++;
		if( data->cur_pos >= mem_get_size( data->text ) )
		{
		    data->text = (uchar*)mem_resize( data->text, mem_get_size( data->text ) + 16 );
		}
	    }
	    draw_window( win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    //Fill window:
	    col2 = col = win->color;
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, col, wm );
	    //Draw text and cursor:
	    xp = 1;
	    for( p = 0; p < mem_get_size( data->text ); p++ )
	    {
		int ty = ( win->ysize - char_y_size( wm ) ) / 2;
		s = data->text[ p ];
		if( s == 0 ) s = ' ';
		if( data->cur_pos == p && data->active ) 
		    col2 = wm->white;
		else
		    col2 = col;
		win_draw_char( win, s, xp, ty, wm->black, col2, wm );
		xp += char_x_size( s, wm );
		if( data->text[ p ] == 0 ) break;
	    }
	    //Draw border:
	    win_draw_frame3d( win, 0, 0, win->xsize, win->ysize, col, 0, wm );
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    if( data->text ) mem_free( data->text );
	    retval = 1;
	    break;
    }
    return retval;
}

void text_set_text( WINDOWPTR win, char *text, window_manager *wm )
{
    if( win )
    {
	text_data *data = (text_data*)win->data;
	if( mem_get_size( data->text ) < mem_strlen( text ) + 1 )
	{
	    //Resize text buffer:
	    data->text = (uchar*)mem_resize( data->text, mem_strlen( text ) + 1 );
	}
	mem_copy( data->text, text, mem_strlen( text ) + 1 );
	draw_window( win, wm );
    }
}

char *text_get_text( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	text_data *data = (text_data*)win->data;
	return (char*)data->text;
    }
    return 0;
}

COLORPTR CREATE_BITMAP_BUTTON = 0;
int CREATE_BITMAP_BUTTON_XSIZE = 0;
int CREATE_BITMAP_BUTTON_YSIZE = 0;
char *CREATE_BUTTON_WITH_TEXT_MENU = 0;

struct button_data
{
    COLORPTR bmp;
    int bmp_xsize;
    int bmp_ysize;
    char *text_menu;
    int pushed;
};

int button_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    button_data *data = (button_data*)win->data;
    COLOR col;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( button_data );
	    break;
	case EVT_AFTERCREATE:
	    data->bmp = CREATE_BITMAP_BUTTON;
	    data->bmp_xsize = CREATE_BITMAP_BUTTON_XSIZE;
	    data->bmp_ysize = CREATE_BITMAP_BUTTON_YSIZE;
	    data->text_menu = CREATE_BUTTON_WITH_TEXT_MENU;
	    data->pushed = 0;
	    CREATE_BITMAP_BUTTON = 0;
	    CREATE_BITMAP_BUTTON_XSIZE = 0;
	    CREATE_BITMAP_BUTTON_YSIZE = 0;
	    CREATE_BUTTON_WITH_TEXT_MENU = 0;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	    if( evt->mouse_button & BUTTON_LEFT )
	    {
		data->pushed = 1;
		draw_window( win, wm );
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( evt->mouse_button & BUTTON_LEFT && data->pushed )
	    {
		if( wm->prev_focus_win && wm->prev_focus_win->visible )
		    wm->focus_win = wm->prev_focus_win;
		if( data->pushed )
		{
		    if( rx >= 0 && rx < win->xsize &&
			ry >= 0 && ry < win->ysize )
		    {
			if( data->text_menu )
			{
			    //Show popup with text menu:
			    win->action_result = popup_menu( win->name, data->text_menu, win->screen_x, win->screen_y, wm->colors[ 12 ], wm );
			}
		    }
		    data->pushed = 0;
		    draw_window( win, wm );
		    if( rx >= 0 && rx < win->xsize &&
			ry >= 0 && ry < win->ysize &&
			win->action_handler )
		    {
			win->action_handler( win->handler_data, win, wm );
		    }
		}
		retval = 1;
	    }
	    break;
	case EVT_DRAW:
	    if( data->pushed )
		col = blend( win->color, 0, 64 );
	    else
		col = win->color;
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, col, wm );
	    if( data->bmp )
	    {
		//Draw bitmap;
		win_draw_bitmap( 
		    win, 
		    ( win->xsize - data->bmp_xsize ) / 2 + data->pushed,
		    ( win->ysize - data->bmp_ysize ) / 2 + data->pushed,
		    data->bmp_xsize,
		    data->bmp_ysize,
		    data->bmp,
		    wm );
	    }
	    else
	    if( win->name )
	    {
		//Draw name:
		int tx = win->xsize / 2;
		int ty = win->ysize / 2;
		if( win->name[ 0 ] == 30 )
		{
		    //UP:
		    win_draw_box( win, tx - 3, ty + 1, 7, 1, wm->black, wm );
		    win_draw_box( win, tx - 2, ty + 0, 5, 1, wm->black, wm );
		    win_draw_box( win, tx - 1, ty - 1, 3, 1, wm->black, wm );
		    win_draw_box( win, tx - 0, ty - 2, 1, 1, wm->black, wm );
		}
		else
		if( win->name[ 0 ] == 31 )
		{
		    //DOWN:
		    win_draw_box( win, tx - 3, ty - 2, 7, 1, wm->black, wm );
		    win_draw_box( win, tx - 2, ty - 1, 5, 1, wm->black, wm );
		    win_draw_box( win, tx - 1, ty - 0, 3, 1, wm->black, wm );
		    win_draw_box( win, tx - 0, ty + 1, 1, 1, wm->black, wm );
		}
		else
		if( win->name[ 0 ] == 17 )
		{
		    //LEFT:
		    win_draw_box( win, tx + 1, ty - 3, 1, 7, wm->black, wm );
		    win_draw_box( win, tx + 0, ty - 2, 1, 5, wm->black, wm );
		    win_draw_box( win, tx - 1, ty - 1, 1, 3, wm->black, wm );
		    win_draw_box( win, tx - 2, ty - 0, 1, 1, wm->black, wm );
		}
		else
		if( win->name[ 0 ] == 16 )
		{
		    //LEFT:
		    win_draw_box( win, tx - 2, ty - 3, 1, 7, wm->black, wm );
		    win_draw_box( win, tx - 1, ty - 2, 1, 5, wm->black, wm );
		    win_draw_box( win, tx - 0, ty - 1, 1, 3, wm->black, wm );
		    win_draw_box( win, tx + 1, ty - 0, 1, 1, wm->black, wm );
		}
		else
		{
		    tx = ( win->xsize - string_size( win->name, wm ) ) / 2;
		    ty = ( win->ysize - char_y_size( wm ) ) / 2;
		    win_draw_string( win, win->name, tx + data->pushed, ty + data->pushed, wm->black, col, wm );
		}
	    }
	    //Draw border:
	    win_draw_frame3d( win, 0, 0, win->xsize, win->ysize, col, 1, wm );
	    retval = 1;
	    break;
    }
    return retval;
}

struct wlist_data
{
    WINDOWPTR this_window;
    list_data *list;
    int prev_selected;
    WINDOWPTR scrollbar;
    int last_action;	//0 - none; 1 - up/down buttons; 2 - mouse click; 3 - enter
};

int list_scrollbar_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    wlist_data *data = (wlist_data*)user_data;
    data->list->start_item = scrollbar_get_value( win, wm );
    if( data->list->start_item < 0 ) data->list->start_item = 0;
    draw_window( data->this_window, wm );
    return 0;
}

int list_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    wlist_data *data = (wlist_data*)win->data;
    COLOR col, col2;
    int i, yp;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    int xoffset = 1;
    int yoffset = 2;
    int ychar_size = char_y_size( wm );
    int ychars = ( win->ysize - xoffset - 1 ) / ychar_size;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( wlist_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->list = (list_data*)MEM_NEW( HEAP_DYNAMIC, sizeof( list_data ) );
	    list_init( data->list );
	    data->prev_selected = -999;
	    //SCROLLBAR:
	    CREATE_VERTICAL_SCROLLBAR = 1;
	    data->scrollbar = new_window( "scroll", 1, 1, 1, 1, wm->colors[ 10 ], win, scrollbar_handler, wm );
	    set_window_controller( data->scrollbar, 0, wm, CPERC, 100, CSUB, 1, CEND );
	    set_window_controller( data->scrollbar, 1, wm, 1, CEND );
	    set_window_controller( data->scrollbar, 2, wm, CPERC, 100, CSUB, SCROLLBAR_SIZE + 1, CEND );
	    set_window_controller( data->scrollbar, 3, wm, CPERC, 100, CSUB, 1, CEND );
	    set_handler( data->scrollbar, list_scrollbar_handler, data, wm );
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    {
		data->last_action = 0;
		int updown = 0;
		int step = 0;
		if( evt->button == KEY_ESCAPE ) { data->last_action = 4; updown = 0; step = 0; }
		if( evt->button == KEY_ENTER ||
		    evt->button == ' ' ) { data->last_action = 2; updown = 0; step = 0; }
		if( evt->button == KEY_UP ) { data->last_action = 1; updown = 1; step = 1; }
		if( evt->button == KEY_PAGEUP ) { data->last_action = 1; updown = 1; step = ychars; }
		if( evt->button == KEY_DOWN ) { data->last_action = 1; updown = 2; step = 1; }
		if( evt->button == KEY_PAGEDOWN ) { data->last_action = 1; updown = 2; step = ychars; }
		if( updown == 1 )
		{
		    data->list->selected_item -= step;
		    if( data->list->selected_item < 0 ) data->list->selected_item = 0;
		    if( data->list->selected_item < data->list->start_item )
			data->list->start_item = data->list->selected_item;
		    draw_window( win, wm );
		}
		if( updown == 2 )
		{
		    data->list->selected_item += step;
		    if( data->list->selected_item >= data->list->items_num ) 
			data->list->selected_item = data->list->items_num - 1;
		    if( data->list->selected_item >= data->list->start_item + ychars )
			data->list->start_item = data->list->selected_item - ychars + 1;
		    if( data->list->start_item < 0 ) data->list->start_item = 0;
		    draw_window( win, wm );
		}
		if( win->action_handler )
		    win->action_handler( win->handler_data, win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    if( evt->mouse_button & BUTTON_LEFT )
	    {
		data->list->selected_item = ( ry - yoffset ) / ychar_size + data->list->start_item;
		if( data->list->selected_item >= data->list->items_num ) 
		    data->list->selected_item = data->list->items_num - 1;
		if( data->prev_selected != data->list->selected_item || evt->event_type == EVT_MOUSEBUTTONDOWN )
		{
		    //Control parameters:
		    //If lower win bound:
		    if( data->list->selected_item < data->list->start_item )
			data->list->start_item = data->list->selected_item;
		    //If higher win bound:
		    if( data->list->selected_item >= data->list->start_item + ychars )
			data->list->start_item = data->list->selected_item - ychars + 1;
		    if( data->list->start_item < 0 ) data->list->start_item = 0;
		    draw_window( win, wm );
		}
		data->prev_selected = data->list->selected_item;
	    }
	    if( evt->mouse_button & BUTTON_SCROLLUP )
	    {
		data->list->start_item -= 2;
		if( data->list->start_item < 0 ) data->list->start_item = 0;
		draw_window( win, wm );
	    }
	    if( evt->mouse_button & BUTTON_SCROLLDOWN )
	    {
		data->list->start_item += 2;
		if( data->list->start_item >= data->list->items_num - ychars ) 
		    data->list->start_item = data->list->items_num - ychars;
		if( data->list->start_item < 0 ) data->list->start_item = 0;
		draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( evt->mouse_button & BUTTON_LEFT )
	    {
		data->last_action = 3;
		if( data->list->selected_item >= 0 && data->list->selected_item < data->list->items_num )
		    if( win->action_handler )
			win->action_handler( win->handler_data, win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    //Set scrollbar parameters:
	    {
		int page_size = ychars;
		scrollbar_set_parameters( data->scrollbar, data->list->start_item, data->list->items_num - page_size, page_size, 1, wm );
	    }
	    //Fill window:
	    col2 = col = win->color;
	    //win_draw_box( win, 1, 1, win->xsize-2, win->ysize-2, col, wm );
	    //Draw items:
	    yp = yoffset;
	    if( yp )
		win_draw_box( win, 1, 1, win->xsize - 2, yp - 1, win->color, wm );
	    for( i = data->list->start_item; i < data->list->items_num; i++ )
	    {
		COLOR fc = wm->black;
		if( list_get_attr( i, data->list ) == 1 ) fc = blend( 0, wm->white, 128 );
		if( data->list->selected_item == i )
		{
		    fc = COLORMASK;
		    col2 = blend( col, wm->black, 64 );
		}
		else
		    col2 = col;
		win_draw_string( win, list_get_item( i, data->list ), xoffset, yp, fc, col2, wm );
		int len = string_size( list_get_item( i, data->list ), wm );
		if( len + xoffset < win->xsize - 1 )
		    win_draw_box( win, len + xoffset, yp, win->xsize - ( len + xoffset ) - 1, char_y_size(wm), col2, wm );
		yp += char_y_size( wm );
	    }
	    if( yp < win->ysize - 1 )
		win_draw_box( win, 1, yp, win->xsize - 2, win->ysize - yp - 1, win->color, wm );
	    //Draw border:
	    win_draw_frame3d( win, 0, 0, win->xsize, win->ysize, col, 0, wm );
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    if( data->list ) { list_close( data->list ); mem_free( data->list ); }
	    retval = 1;
	    break;
    }
    return retval;
}

list_data *list_get_data( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	wlist_data *data = (wlist_data*)win->data;
	return data->list;
    }
    return 0;
}

int CREATE_VERTICAL_SCROLLBAR = 0;
int CREATE_REVERSE_SCROLLBAR = 0;
char text_up[ 2 ] = { 30, 0 };
char text_down[ 2 ] = { 31, 0 };
char text_left[ 2 ] = { 17, 0 };
char text_right[ 2 ] = { 16, 0 };

struct scrollbar_data
{
    WINDOWPTR this_window;
    WINDOWPTR but1;
    WINDOWPTR but2;
    int vert;
    int rev;
    int max_value;
    int page_size;
    int step_size;
    int cur;
    int bar_selected;
    int drag_start;
    int drag_start_val;
};

int scrollbar_dec_button( void *user_data, WINDOWPTR win, window_manager *wm )
{
    scrollbar_data *data = (scrollbar_data*)user_data;
    data->cur -= data->step_size;
    if( data->cur < 0 ) data->cur = 0;
    if( data->cur > data->max_value ) data->cur = data->max_value;
    if( data->this_window->action_handler )
	data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    draw_window( data->this_window, wm );
    return 0;
}

int scrollbar_inc_button( void *user_data, WINDOWPTR win, window_manager *wm )
{
    scrollbar_data *data = (scrollbar_data*)user_data;
    data->cur += data->step_size;
    if( data->cur < 0 ) data->cur = 0;
    if( data->cur > data->max_value ) data->cur = data->max_value;
    if( data->this_window->action_handler )
	data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    draw_window( data->this_window, wm );
    return 0;
}

int scrollbar_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    scrollbar_data *data = (scrollbar_data*)win->data;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    int button_size = SCROLLBAR_SIZE;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( scrollbar_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    if( CREATE_VERTICAL_SCROLLBAR )
	    {
		data->vert = 1;
		data->but1 = new_window( text_up, 0, 0, win->xsize, button_size, win->color, win, button_handler, wm );
		data->but2 = new_window( text_down, 0, win->ysize - button_size, win->xsize, button_size, win->color, win, button_handler, wm );
		set_window_controller( data->but1, 0, wm, CPERC, 0, CEND );
		set_window_controller( data->but1, 1, wm, CPERC, 0, CEND );
		set_window_controller( data->but1, 2, wm, CPERC, 100, CEND );
		set_window_controller( data->but1, 3, wm, button_size, CEND );
		set_window_controller( data->but2, 0, wm, CPERC, 0, CEND );
		set_window_controller( data->but2, 1, wm, CPERC, 100, CEND );
		set_window_controller( data->but2, 2, wm, CPERC, 100, CEND );
		set_window_controller( data->but2, 3, wm, CPERC, 100, CSUB, button_size, CEND );
		if( CREATE_REVERSE_SCROLLBAR )
		{
		    set_handler( data->but1, scrollbar_inc_button, data, wm );
		    set_handler( data->but2, scrollbar_dec_button, data, wm );
		}
		else
		{
		    set_handler( data->but1, scrollbar_dec_button, data, wm );
		    set_handler( data->but2, scrollbar_inc_button, data, wm );
		}
	    }
	    else
	    {
		data->vert = 0;
		data->but1 = new_window( text_right, 0, 0, button_size, win->ysize, win->color, win, button_handler, wm );
		data->but2 = new_window( text_left, win->xsize - button_size, 0, button_size, win->ysize, win->color, win, button_handler, wm );
		set_window_controller( data->but1, 0, wm, CPERC, 100, CEND );
		set_window_controller( data->but1, 1, wm, CPERC, 0, CEND );
		set_window_controller( data->but1, 2, wm, CPERC, 100, CSUB, button_size, CEND );
		set_window_controller( data->but1, 3, wm, CPERC, 100, CEND );
		set_window_controller( data->but2, 0, wm, CPERC, 0, CEND );
		set_window_controller( data->but2, 1, wm, CPERC, 0, CEND );
		set_window_controller( data->but2, 2, wm, button_size, CEND );
		set_window_controller( data->but2, 3, wm, CPERC, 100, CEND );
		if( CREATE_REVERSE_SCROLLBAR )
		{
		    set_handler( data->but1, scrollbar_dec_button, data, wm );
		    set_handler( data->but2, scrollbar_inc_button, data, wm );
		}
		else
		{
		    set_handler( data->but1, scrollbar_inc_button, data, wm );
		    set_handler( data->but2, scrollbar_dec_button, data, wm );
		}
	    }
	    data->rev = CREATE_REVERSE_SCROLLBAR;
	    data->cur = 0;
	    data->max_value = 0;
	    data->page_size = 0;
	    data->step_size = 1;
	    data->bar_selected = 0;
	    CREATE_VERTICAL_SCROLLBAR = 0;
	    CREATE_REVERSE_SCROLLBAR = 0;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    {
		int working_area = 0;
		if( data->vert )
		    working_area = win->ysize - button_size * 2;
		else
		    working_area = win->xsize - button_size * 2;
		int coef = 0;
		if( ( data->max_value + data->page_size ) ) 
		    coef = ( working_area << 11 ) / ( data->max_value + data->page_size );
		int coef2 = 0;
		if( working_area ) coef2 = ( ( data->max_value + data->page_size ) << 11 ) / working_area;
		int pos;
		if( data->rev )
		    pos = ( ( data->max_value - data->cur ) * coef ) >> 11;
		else
		    pos = ( data->cur * coef ) >> 11;
		int bar_size = ( data->page_size * coef ) >> 11;
		if( bar_size < 2 ) bar_size = 2;
		if( bar_size > working_area ) { pos = 0; bar_size = working_area; }
		if( evt->event_type == EVT_MOUSEBUTTONDOWN && evt->mouse_button & BUTTON_LEFT )
		{
		    if( data->vert )
		    {
			if( ry >= button_size + pos && ry < button_size + pos + bar_size )
			{
			    data->bar_selected = 1;
			    data->drag_start = ry;
			    data->drag_start_val = data->cur;
			}
			int page = data->page_size;
			if( data->rev ) page = -page;
			if( ry < button_size + pos )
			    data->cur -= page;
			if( ry >= button_size + pos + bar_size )
			    data->cur += page;
		    }
		    else
		    {
			if( rx >= button_size + pos && rx < button_size + pos + bar_size )
			{
			    data->bar_selected = 1;
			    data->drag_start = rx;
			    data->drag_start_val = data->cur;
			}
			int page = data->page_size;
			if( data->rev ) page = -page;
			if( rx < button_size + pos )
			    data->cur -= page;
			if( rx >= button_size + pos + bar_size )
			    data->cur += page;
		    }
		}
		else
		{
		    //Move:
		    if( data->bar_selected )
		    {
			int d = 0;
			if( data->vert ) 
			    d = ry - data->drag_start;
			else
			    d = rx - data->drag_start;
			if( data->rev )
			    data->cur = data->drag_start_val - ( ( d * coef2 ) >> 11 );
			else
			    data->cur = data->drag_start_val + ( ( d * coef2 ) >> 11 );
		    }
		}
		//Bounds control:
		if( data->cur < 0 ) data->cur = 0;
		if( data->cur > data->max_value ) data->cur = data->max_value;
		//User handler:
		if( win->action_handler )
		    win->action_handler( win->handler_data, win, wm );
		//Redraw it:
		draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( wm->prev_focus_win && wm->prev_focus_win->visible )
		wm->focus_win = wm->prev_focus_win;
	    data->bar_selected = 0;
	    draw_window( win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		int working_area = 0;
		if( data->vert )
		    working_area = win->ysize - button_size * 2;
		else
		    working_area = win->xsize - button_size * 2;
		int coef = 0;
		if( ( data->max_value + data->page_size ) ) 
		    coef = ( working_area << 11 ) / ( data->max_value + data->page_size );
		int pos;
		if( data->rev )
		    pos = ( ( data->max_value - data->cur ) * coef ) >> 11;
		else
		    pos = ( data->cur * coef ) >> 11;
		int bar_size = ( data->page_size * coef ) >> 11;
		if( bar_size < 2 ) bar_size = 2;
		if( bar_size > working_area ) { pos = 0; bar_size = working_area; }
		COLOR lcol = blend( win->color, COLORMASK, 64 );
		COLOR dcol = blend( win->color, 0, 64 );
		COLOR bar_color = win->color;
		if( data->bar_selected ) bar_color = blend( bar_color, wm->black, 64 );
		if( data->vert )
		{
		    win_draw_box( win, 0, button_size, win->xsize, button_size + pos, blend( win->color, wm->black, 40 ), wm );
		    win_draw_box( win, 1, button_size + pos + 1, win->xsize - 2, bar_size - 2, bar_color, wm );
		    win_draw_frame3d( win, 0, button_size + pos, win->xsize, bar_size, bar_color, 1, wm );
		    win_draw_box( win, 0, button_size + pos + bar_size, win->xsize, working_area - (pos + bar_size), blend( win->color, wm->black, 40 ), wm );
		}
		else
		{
		    win_draw_box( win, button_size, 0, button_size + pos, win->ysize, blend( win->color, wm->black, 40 ), wm );
		    win_draw_box( win, button_size + pos + 1, 1, bar_size - 2, win->ysize - 2, bar_color, wm );
		    win_draw_frame3d( win, button_size + pos, 0, bar_size, win->ysize, bar_color, 1, wm );
		    win_draw_box( win, button_size + pos + bar_size, 0, working_area - (pos + bar_size), win->ysize, blend( win->color, wm->black, 40 ), wm );
		}
	    }
	    retval = 1;
	    break;
    }
    return retval;
}

void scrollbar_set_parameters( WINDOWPTR win, int cur, int max_value, int page_size, int step_size, window_manager *wm )
{
    if( win )
    {
	scrollbar_data *data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->cur = cur;
	    data->max_value = max_value;
	    data->page_size = page_size;
	    data->step_size = step_size;
	}
    }
}

void scrollbar_set_value( WINDOWPTR win, int val, window_manager *wm )
{
    if( win )
    {
	scrollbar_data *data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->cur = val;
	    if( data->cur < 0 ) data->cur = 0;
	    if( data->cur > data->max_value ) data->cur = data->max_value;
	}
    }
}

int scrollbar_get_value( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	scrollbar_data *data = (scrollbar_data*)win->data;
	if( data )
	{
	    return data->cur;
	}
    }
    return 0;
}

char *FILES_PROPS = 0;
char *FILES_MASK = 0;
char *FILES_RESULTED_FILENAME = 0;

struct files_data
{
    WINDOWPTR this_window;
    WINDOWPTR go_up_button;
    WINDOWPTR disk_buttons[ 32 ];
    WINDOWPTR path;
    WINDOWPTR name;
    WINDOWPTR files;
    WINDOWPTR ok_button;
    WINDOWPTR cancel_button;
    char *mask;
    char *props_file;		    //File with dialog properties
    char *prop_path[ 32 ];
    int prop_cur_file[ 32 ];
    int	prop_cur_disk;
    char *resulted_filename;
};

int files_ok_button_handler( void *user_data, WINDOWPTR win, window_manager *wm );
int files_cancel_button_handler( void *user_data, WINDOWPTR win, window_manager *wm );
void files_refresh_list( WINDOWPTR win, window_manager *wm );
void files_save_props( WINDOWPTR win, window_manager *wm );

int files_list_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    wlist_data *wdata = (wlist_data*)win->data;
    files_data *data = (files_data*)user_data;
    list_data *ldata = list_get_data( data->files, wm );
    if( ldata && wdata->last_action )
    {
	if( wdata->last_action == 4 )
	{
	    //ESCAPE KEY:
	    files_cancel_button_handler( data, data->ok_button, wm );
	    return 0;
	}
	char *item = list_get_item( list_get_selected_num( ldata ), ldata );
	int subdir = 0;
	if( item )
	{
	    if( wdata->last_action == 1 )
	    {
		//UP/DOWN KEY:
		text_set_text( data->name, item, wm );
	    }
	    if( wdata->last_action == 2 )
	    {
		//ENTER/SPACE KEY:
		if( list_get_attr( list_get_selected_num( ldata ), ldata ) == 1 )
		{
		    //It's a dir:
		    subdir = 1;
		}
		else
		{
		    //Select file:
		    files_ok_button_handler( data, data->ok_button, wm );
		    return 0;
		}
	    }
	    if( wdata->last_action == 3 )
	    {
		//MOUSE CLICK:
		if( list_get_attr( list_get_selected_num( ldata ), ldata ) == 1 )
		{
		    //It's a dir:
		    subdir = 1;
		}
		else
		{
		    text_set_text( data->name, item, wm );
		}
	    }
	    if( subdir )
	    {
		//Go to the subdir:
		char *path = data->prop_path[ data->prop_cur_disk ];
		if( path )
		    path = (char*)mem_resize( path, mem_strlen( path ) + mem_strlen( item ) + 2 );
		else
		{
		    path = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( item ) + 2 );
		    path[ 0 ] = 0;
		}
		mem_strcat( path, item );
		mem_strcat( path, "/" );
		data->prop_path[ data->prop_cur_disk ] = path;
		data->prop_cur_file[ data->prop_cur_disk ] = -1;
		text_set_text( data->path, path, wm );
		files_refresh_list( data->this_window, wm );
		draw_window( data->this_window, wm );
	    }
	}
	data->prop_cur_file[ data->prop_cur_disk ] = ldata->selected_item;
    }
    return 0;
}

int files_disk_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)user_data;
    
    if( win == data->go_up_button )
    {
	//Go to parent dir:
	char *path = data->prop_path[ data->prop_cur_disk ];
	if( path )
	{
	    int p;
	    for( p = mem_strlen( path ) - 2; p >= 0; p-- )
	    {
		if( path[ p ] == '/' ) { path[ p + 1 ] = 0; break; }
	    }
	    if( p < 0 ) path[ 0 ] = 0;
	}
	data->prop_cur_file[ data->prop_cur_disk ] = -1;
    }
    else
    {
	for( int a = 0; a < disks; a++ )
	{
	    data->disk_buttons[ a ]->color = wm->colors[ 12 ];
	    if( data->disk_buttons[ a ] == win )
		data->prop_cur_disk = a;
	}
	//Show selected disk:
	data->disk_buttons[ data->prop_cur_disk ]->color = wm->colors[ 15 ];
    }

    //Set new filelist:
    files_refresh_list( data->this_window, wm );
    //Set new path:
    if( data->prop_path[ data->prop_cur_disk ] )
	text_set_text( data->path, data->prop_path[ data->prop_cur_disk ], wm );
    else
	text_set_text( data->path, "", wm );

    //Save props:
    files_save_props( data->this_window, wm );
    draw_window( data->this_window, wm );
    return 0;
}

int files_ok_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)user_data;
    if( data->resulted_filename )
    {
	char *res = data->resulted_filename;
	res[ 0 ] = 0;
	mem_strcat( res, get_disk_name( data->prop_cur_disk ) );
	if( data->prop_path[ data->prop_cur_disk ] )
	    mem_strcat( res, data->prop_path[ data->prop_cur_disk ] );
	char *fname = text_get_text( data->name, wm );
	if( fname )
	    mem_strcat( res, fname );
    }
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int files_cancel_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)user_data;
    if( data->resulted_filename )
    {
	char *res = data->resulted_filename;
	res[ 0 ] = 0;
    }
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int files_name_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)user_data;
    char *name = text_get_text( data->name, wm );
    if( name && data->mask )
    {
	int a = 0;
	for( a = mem_strlen( name ) - 1; a >= 0; a-- )
	    if( name[ a ] == '.' ) break;
	if( a < 0 ) 
	{
	    //Dot not found:
	    char *new_name = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( name ) + 10 );
	    new_name[ 0 ] = 0;
	    mem_strcat( new_name, name );
	    a = mem_strlen( name );
	    new_name[ a ] = '.';
	    a++;
	    int p = 0;
	    for(;;)
	    {
		new_name[ a ] = data->mask[ p ];
		if( new_name[ a ] == '/' ||
		    new_name[ a ] == 0 )
		{
		    new_name[ a ] = 0;
		    break;
		}
		p++;
		a++;
	    }
	    text_set_text( data->name, new_name, wm );
	    draw_window( data->name, wm );
	    mem_free( new_name );
	}
	else
	{
	    //Dot found:
	    char *new_name = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( name ) + 10 );
	    new_name[ 0 ] = 0;
	    mem_strcat( new_name, name );
	    a++;
	    int p = 0;
	    for(;;)
	    {
		new_name[ a ] = data->mask[ p ];
		if( new_name[ a ] == '/' ||
		    new_name[ a ] == 0 )
		{
		    new_name[ a ] = 0;
		    break;
		}
		p++;
		a++;
	    }
	    text_set_text( data->name, new_name, wm );
	    draw_window( data->name, wm );
	    mem_free( new_name );
	}
    }
    return 0;
}

void files_refresh_list( WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)win->data;
    list_data *ldata = list_get_data( data->files, wm );
    list_close( ldata );
    list_init( ldata );
    find_struct fs;
    char *disk_name = get_disk_name( data->prop_cur_disk );
    char *path = data->prop_path[ data->prop_cur_disk ];
    char *res = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( disk_name ) + mem_strlen( path ) + 1 );
    res[ 0 ] = 0;
    mem_strcat( res, disk_name );
    mem_strcat( res, path );
    fs.start_dir = res;
    fs.mask = data->mask;
    if( find_first( &fs ) )
    {
	if( fs.name[ 0 ] != '.' )
	{
	    list_add_item( fs.name, fs.type, ldata );
	}
    }
    while( find_next( &fs ) )
    {
	if( fs.name[ 0 ] != '.' )
	{
	    list_add_item( fs.name, fs.type, ldata );
	}
    }
    find_close( &fs );
    mem_free( res );
    list_sort( ldata );

    //Set file num:
    list_set_selected_num( data->prop_cur_file[ data->prop_cur_disk ], ldata );
}

void files_load_string( FILE *f, char *dest )
{
    for( int p = 0; p < 4096; p++ )
    {
	if( feof( f ) ) { dest[ p ] = 0; break; }
	dest[ p ] = getc( f );
	if( dest[ p ] == 0 ) break;
    }
}

int files_get_disk_num( char *name )
{
    for( int a = 0; a < disks; a++ )
    {
	if( mem_strcmp( name, get_disk_name( a ) ) == 0 ) return a;
    }
    return -1;
}

void files_load_props( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	files_data *data = (files_data*)win->data;
	if( data->props_file )
	{
	    FILE *f = fopen( data->props_file, "rb" );
	    if( f )
	    {
		char *temp_str = (char*)MEM_NEW( HEAP_DYNAMIC, 4096 );

		//Current disk:
		files_load_string( f, temp_str );
		int d = files_get_disk_num( temp_str );
		if( d == -1 ) 
		    data->prop_cur_disk = get_current_disk();
		else
		    data->prop_cur_disk = d;

		//Disks info:
		for( int a = 0; a < 32; a++ )
		{
		    if( feof( f ) ) break;
		    //Disk name:
		    files_load_string( f, temp_str );
		    int d = files_get_disk_num( temp_str );
		    if( d != -1 )
		    {
			//Path:
			files_load_string( f, temp_str );
			if( data->prop_path[ d ] )
			{
			    //Remove old path:
			    mem_free( data->prop_path[ d ] );
			    data->prop_path[ d ] = 0;
			}
			if( temp_str[ 0 ] )
			{
			    //Set new path:
			    data->prop_path[ d ] = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( temp_str ) + 1 );
			    mem_copy( data->prop_path[ d ], temp_str, mem_strlen( temp_str ) + 1 );
			}
		    }
		    int file_num;
		    fread( &file_num, 4, 1, f );
		    if( d != -1 )
			data->prop_cur_file[ d ] = file_num;
		}
		
		mem_free( temp_str );
		fclose( f );
	    }
	    else
	    {
		//File not found
		for( int a = 0; a < 32; a++ )
		{
		    data->prop_path[ a ] = 0;
		    data->prop_cur_file[ a ] = 0;
		}
		//Set current disk:
		data->prop_cur_disk = get_current_disk();
		//Set current path:
		char *cdir = get_current_dir();
		data->prop_path[ data->prop_cur_disk ] = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( cdir ) + 1 );
		mem_copy( data->prop_path[ data->prop_cur_disk ], cdir, mem_strlen( cdir ) + 1 );
	    }
	}
	//Refresh list:
	files_refresh_list( win, wm );
    }
}

void files_save_props( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	files_data *data = (files_data*)win->data;
	if( data->props_file )
	{
	    FILE *f = fopen( data->props_file, "wb" );
	    if( f )
	    {
		//Save current disk:
		char *path;
		char *disk_name = get_disk_name( data->prop_cur_disk );
	        fwrite( disk_name, 1, mem_strlen( disk_name ) + 1, f );
		//Save other properties:
		for( int a = 0; a < disks; a++ )
		{
		    //Disk name:
		    disk_name = get_disk_name( a );
		    fwrite( disk_name, 1, mem_strlen( disk_name ) + 1, f );
		    //Current path:
		    if( data->prop_path[ a ] == 0 )
			fputc( 0, f );
		    else
		    {
			path = data->prop_path[ a ];
			fwrite( path, 1, mem_strlen( path ) + 1, f );
		    }
		    //Selected file num:
		    fwrite( &data->prop_cur_file[ a ], 4, 1, f );
		}
		fclose( f );
	    }
	}
    }
}

int files_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    files_data *data = (files_data*)win->data;
    int a, b;
    int but_xsize = 30;
    int but_ysize = char_y_size( wm ) + 8;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( files_data );
	    break;
	case EVT_AFTERCREATE:
	    //DATA INIT:
	    for( a = 0; a < 32; a ++ )
	    {
		data->prop_path[ a ] = 0;
		data->prop_cur_file[ a ] = 0;
		data->prop_cur_disk = get_current_disk();
	    }
	    data->props_file = FILES_PROPS;
	    data->mask = FILES_MASK;
	    data->resulted_filename = FILES_RESULTED_FILENAME;
	    data->this_window = win;

	    //DISKS:
	    for( a = 0; a < 32; a++ ) data->disk_buttons[ a ] = 0;
	    get_disks();
	    data->go_up_button = new_window( 
		"..", 
		2, 2, but_xsize, but_ysize, 
		wm->colors[ 12 ],
		win,
		button_handler, 
		wm );
	    set_handler( data->go_up_button, files_disk_button_handler, data, wm );
	    for( a = 0; a < disks; a++ )
	    {
		data->disk_buttons[ a ] = new_window( 
		    get_disk_name( a ), 
		    2 + ( but_xsize + 4 ) * ( a + 1 ), 2, but_xsize, but_ysize, 
		    wm->colors[ 12 ],
		    win,
		    button_handler, 
		    wm );
		set_handler( data->disk_buttons[ a ], files_disk_button_handler, data, wm );
	    }

	    //DIRECTORY:
	    a = string_size( "#####", wm ) + 4;
	    b = but_ysize + 4;
	    new_window( "Path:", 2, b, a, char_y_size(wm)+4, 0, win, label_handler, wm );
	    data->path = new_window( "pathn", 0, 0, 1, 1, wm->colors[ 12 ], win, text_handler, wm );
	    set_window_controller( data->path, 0, wm, 2 + a, CEND );
	    set_window_controller( data->path, 1, wm, b, CEND );
	    set_window_controller( data->path, 2, wm, CPERC, 100, CEND );
	    set_window_controller( data->path, 3, wm, b + char_y_size( wm ) + 4, CEND );

	    //FILENAME:
	    b += char_y_size( wm ) + 4 + 2;
	    new_window( "Name:", 2, b, string_size("Name:",wm)+4, char_y_size(wm)+4, 0, win, label_handler, wm );
	    data->name = new_window( "namen", 0, 0, 1, 1, wm->colors[ 12 ], win, text_handler, wm );
	    set_handler( data->name, files_name_handler, data, wm );
	    set_window_controller( data->name, 0, wm, 2 + a, CEND );
	    set_window_controller( data->name, 1, wm, b, CEND );
	    set_window_controller( data->name, 2, wm, CPERC, 100, CEND );
	    set_window_controller( data->name, 3, wm, b + char_y_size( wm ) + 4, CEND );

	    //FILES:
	    b += char_y_size( wm ) + 4 + 3;
	    data->files = new_window( "files", 0, b, 100, 100, wm->colors[ 12 ], win, list_handler, wm );
	    set_handler( data->files, files_list_handler, data, wm );
	    set_window_controller( data->files, 0, wm, CPERC, 0, CADD, 2, CEND );
	    set_window_controller( data->files, 1, wm, b, CEND );
	    set_window_controller( data->files, 2, wm, CPERC, 100, CSUB, 2, CEND );
	    set_window_controller( data->files, 3, wm, CPERC, 100, CSUB, but_ysize + 4 + 6, CEND );

	    //BUTTONS:
	    data->ok_button = new_window( "OK", 0, 0, 1, 1, wm->colors[ 12 ], win, button_handler, wm );
	    data->cancel_button = new_window( "CANCEL", 0, 0, 1, 1, wm->colors[ 12 ], win, button_handler, wm );
	    set_handler( data->ok_button, files_ok_button_handler, data, wm );
	    set_handler( data->cancel_button, files_cancel_button_handler, data, wm );
	    set_window_controller( data->ok_button, 0, wm, 2, CEND );
	    set_window_controller( data->ok_button, 1, wm, CPERC, 100, CSUB, 2, CEND );
	    set_window_controller( data->ok_button, 2, wm, 2 + 70, CEND );
	    set_window_controller( data->ok_button, 3, wm, CPERC, 100, CSUB, but_ysize + 4 + 2, CEND );
	    set_window_controller( data->cancel_button, 0, wm, 2 + 70 + 1, CEND );
	    set_window_controller( data->cancel_button, 1, wm, CPERC, 100, CSUB, 2, CEND );
	    set_window_controller( data->cancel_button, 2, wm, 2 + 70 + 1 + 70, CEND );
	    set_window_controller( data->cancel_button, 3, wm, CPERC, 100, CSUB, but_ysize + 4 + 2, CEND );

	    //LOAD LAST PROPERTIES:
    	    files_load_props( win, wm );
	    if( data->prop_path[ data->prop_cur_disk ] )
		text_set_text( data->path, data->prop_path[ data->prop_cur_disk ], wm );
	    if( data->prop_cur_file >= 0 )
	    {
		list_data *ldata = list_get_data( data->files, wm );
		char *item = list_get_item( list_get_selected_num( ldata ), ldata );
		if( item ) 
		    text_set_text( data->name, item, wm );
	    }

	    //SHOW CURRENT DISK:
	    data->disk_buttons[ data->prop_cur_disk ]->color = wm->colors[ 15 ];

	    //SET FOCUS:
	    wm->focus_win = data->files;

	    FILES_PROPS = 0;
	    FILES_MASK = 0;
	    FILES_RESULTED_FILENAME = 0;
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    files_save_props( win, wm );
	    for( a = 0; a < 32; a++ )
	    {
		if( data->prop_path[ a ] ) mem_free( data->prop_path[ a ] );
	    }
	    break;
	case EVT_DRAW:
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	    retval = 1;
	    break;
    }
    return retval;
}

char *DIALOG_OK_TEXT = 0;
char *DIALOG_CANCEL_TEXT = 0;
char *DIALOG_TEXT = 0;

struct dialog_data
{
    WINDOWPTR this_window;
    char *text;
};

int dialog_ok_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    dialog_data *data = (dialog_data*)user_data;
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int dialog_cancel_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    dialog_data *data = (dialog_data*)user_data;
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int dialog_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    dialog_data *data = (dialog_data*)win->data;
    int but_ysize = char_y_size( wm ) + 8;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( dialog_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		data->text = DIALOG_TEXT;
		WINDOWPTR w_ok = new_window( DIALOG_OK_TEXT, 2, win->ysize - but_ysize - 6, 70, but_ysize + 4, wm->colors[ 12 ], win, button_handler, wm );
		WINDOWPTR w_cancel = new_window( DIALOG_CANCEL_TEXT, 71, win->ysize - but_ysize - 6, 70, but_ysize + 4, wm->colors[ 12 ], win, button_handler, wm );
		set_handler( w_ok, dialog_ok_button_handler, data, wm );
		set_handler( w_cancel, dialog_cancel_button_handler, data, wm );
		set_window_controller( w_ok, 0, wm, 2, CEND );
		set_window_controller( w_ok, 1, wm, CPERC, 100, CSUB, 2, CEND );
		set_window_controller( w_ok, 2, wm, 2 + 70, CEND );
		set_window_controller( w_ok, 3, wm, CPERC, 100, CSUB, but_ysize + 4 + 2, CEND );
		set_window_controller( w_cancel, 0, wm, 2 + 70 + 1, CEND );
		set_window_controller( w_cancel, 1, wm, CPERC, 100, CSUB, 2, CEND );
		set_window_controller( w_cancel, 2, wm, 2 + 70 + 1 + 70, CEND );
		set_window_controller( w_cancel, 3, wm, CPERC, 100, CSUB, but_ysize + 4 + 2, CEND );
		DIALOG_OK_TEXT = 0;
		DIALOG_CANCEL_TEXT = 0;
		DIALOG_TEXT = 0;

		//SET FOCUS:
		wm->focus_win = win;
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
	    if( data->text )
		win_draw_string( win, data->text, 2, 2, wm->black, win->color, wm );
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( evt->button == KEY_ESCAPE )
	    {
		dialog_cancel_button_handler( data, win, wm );
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	    retval = 1;
	    break;
    }
    return retval;
}

struct popup_data
{
    WINDOWPTR this_window;
    char *text;
    int lines[ 128 ];
    int lines_num;
    int current_selected;
    WINDOWPTR prev_focus;
};

char *CREATE_TEXT_POPUP = 0;

int popup_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->event_win;
    popup_data *data = (popup_data*)win->data;
    int popup_border = 3;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->event_type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( popup_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		//Set focus:
		data->prev_focus = wm->focus_win;
		wm->focus_win = win;

		//Data init:
	    	data->text = 0;
		data->current_selected = -1;
		win->action_result = -1;

		//Set window size:
		int xsize = 0;
		int ysize = 0;
		int x = win->x;
		int y = win->y;
		if( win->name && win->name[ 0 ] != 0 )
		{
		    xsize = string_size( win->name, wm );
		    ysize = char_y_size( wm ) + popup_border;
		}
		if( CREATE_TEXT_POPUP )
		{
		    //Text popup:
		    data->text = (char*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( CREATE_TEXT_POPUP ) + 1 );
		    mem_copy( data->text, CREATE_TEXT_POPUP, mem_strlen( CREATE_TEXT_POPUP ) + 1 );
		    int lines = 1;
		    int max_x_size = 1;
		    int cur_x_size = 0;
		    data->lines[ 0 ] = 0;
		    int nl_code = 0;
		    int text_size = mem_strlen( data->text );
		    for( int i = 0; i < text_size; i++ )
		    {
			if( data->text[ i ] == 0xA ) { data->text[ i ] = 0; lines++; cur_x_size = 0; nl_code = 1; }
			else
			{
			    if( data->text[ i ] != 0xA && data->text[ i ] != 0xD ) 
			    {
				if( nl_code == 1 ) data->lines[ lines - 1 ] = i;
				nl_code = 0;
				cur_x_size += char_x_size( data->text[ i ], wm );
				if( cur_x_size > max_x_size ) max_x_size = cur_x_size;
			    }
			    else
			    {
				data->text[ i ] = 0;
				nl_code = 1;
			    }
			}
		    }
		    data->lines_num = lines;
		    if( max_x_size > xsize ) xsize = max_x_size;
		    ysize += lines * char_y_size( wm );
		}
		xsize += popup_border * 2;
		ysize += popup_border * 2;

		//Control window position:
		if( x + xsize > win->parent->xsize && x > 0 ) x -= ( x + xsize ) - win->parent->xsize;
		if( y + ysize > win->parent->ysize && y > 0 ) y -= ( y + ysize ) - win->parent->ysize;
		if( x < 0 ) x = 0;
		if( y < 0 ) y = 0;

		win->x = x;
		win->y = y;
		win->xsize = xsize;
		win->ysize = ysize;

		CREATE_TEXT_POPUP = 0;
    	    
		retval = 1;
	    }
	    break;
	case EVT_BEFORECLOSE:
	    wm->focus_win = data->prev_focus;
	    if( data->text ) mem_free( data->text );
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    {
		int cur_y = popup_border;
		if( win->name && win->name[ 0 ] != 0 )
		{
		    cur_y += char_y_size( wm ) + popup_border;
		}
		if( data->text && data->lines_num > 0 )
		{
		    data->current_selected = ( ry - cur_y ) / char_y_size( wm );
		}
		//If out of bounds:
		if( rx < 0 || rx >= win->xsize ||
		    ry < cur_y || ry >= win->ysize )
		    data->current_selected = -1;
	    }
	    draw_window( win, wm );
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( data->current_selected >= 0 )
	    {
		//Successful selection:
		win->action_result = data->current_selected;
		remove_window( win, wm );
		recalc_regions( wm );
		draw_window( wm->root_win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_UNFOCUS:
	    data->prev_focus = wm->focus_win; //New window was selected. It must be focused after popup window closing.
	    remove_window( win, wm );
	    recalc_regions( wm );
	    draw_window( wm->root_win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		int cur_y = popup_border;
		win_draw_box( win, 1, 1, win->xsize - 2, win->ysize - 2, win->color, wm );

		//Draw name:
		if( win->name && win->name[ 0 ] != 0 )
		{
		    win_draw_box( win, popup_border - 1, cur_y - 1, win->xsize - popup_border * 2 + 2, char_y_size( wm ) + 2, blend( win->color, wm->black, 64 ), wm );
		    win_draw_string( win, win->name, popup_border, cur_y, wm->white, blend( win->color, wm->black, 64 ), wm );
		    cur_y += char_y_size( wm ) + popup_border;
		}

		//Draw text:
		if( data->text && data->lines_num > 0 )
		{
		    for( int i = 0; i < data->lines_num; i++ )
		    {
			if( i == data->current_selected )
			{
			    win_draw_box( win, popup_border - 1, cur_y, win->xsize - popup_border * 2 + 2, char_y_size( wm ), wm->black, wm );
			    win_draw_string( win, data->text + data->lines[ i ], popup_border, cur_y, wm->white, wm->black, wm );
			}
			else
			{
			    win_draw_string( win, data->text + data->lines[ i ], popup_border, cur_y, wm->black, win->color, wm );
			}
			cur_y += char_y_size( wm );
		    }
		}

		win_draw_frame( win, 0, 0, win->xsize, win->ysize, wm->black, wm );

		retval = 1;
	    }
	    break;
    }
    return retval;
}

int popup_menu( char* name, char* items, int x, int y, COLOR color, window_manager *wm )
{
    CREATE_TEXT_POPUP = items;
    WINDOWPTR win = new_window( name, x, y, 1, 1, color, wm->root_win, popup_handler, wm );
    show_window( win, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    while( 1 )
    {
	FIRST_PART_OF_THE_EVENT_LOOP( wm )
	}
	LAST_PART_OF_THE_EVENT_LOOP( wm )
	if( win->visible == 0 ) break;
    }
    return win->action_result;
}

