#ifdef DIRECTX
    #define XLINESIZE	    g_lPitch
#else
    #define XLINESIZE	    wm->screen_xsize
#endif

void device_draw_pixel( int x, int y, COLOR color, window_manager *wm )
{
    if( !framebuffer ) return;
    framebuffer[ y * XLINESIZE + x ] = color;
}

COLOR device_get_pixel( int x, int y, window_manager *wm )
{
    if( !framebuffer ) return 0;
    return framebuffer[ y * XLINESIZE + x ];
}

void device_draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm )
{
    if( !framebuffer ) return;

    int len_x = x2 - x1; if( len_x < 0 ) len_x = -len_x;
    int len_y = y2 - y1; if( len_y < 0 ) len_y = -len_y;
    int ptr = y1 * XLINESIZE + x1;
    int delta;
    int v = 0, old_v = 0;
    if( len_x > len_y )
    {
	//Horisontal:
	if( len_x != 0 )
	    delta = ( len_y << 10 ) / len_x;
	else
	    delta = 0;
	for( int a = 0; a <= len_x; a++ )
	{
	    framebuffer[ ptr ] = color;
	    old_v = v;
	    v += delta;
	    if( x2 - x1 > 0 ) ptr++; else ptr--;
	    if( ( old_v >> 10 ) != ( v >> 10 ) ) 
	    {
		if( y2 - y1 > 0 )
		    ptr += XLINESIZE;
		else
		    ptr -= XLINESIZE;
	    }
	}
    }
    else
    {
	//Vertical:
	if( len_y != 0 ) 
	    delta = ( len_x << 10 ) / len_y;
	else
	    delta = 0;
	for( int a = 0; a <= len_y; a++ )
	{
	    framebuffer[ ptr ] = color;
	    old_v = v;
	    v += delta;
	    if( y2 - y1 > 0 ) 
		ptr += XLINESIZE;
	    else
		ptr -= XLINESIZE;
	    if( ( old_v >> 10 ) != ( v >> 10 ) ) 
	    {
		if( x2 - x1 > 0 ) ptr++; else ptr--;
	    }
	}
    }
}

void device_draw_box( int x, int y, int xsize, int ysize, COLOR color, window_manager *wm )
{
    if( !framebuffer ) return;

    int ptr = y * XLINESIZE + x;
    for( int cy = 0; cy < ysize; cy++ )
    {
	for( int cx = 0; cx < xsize; cx++ )
	{
	    framebuffer[ ptr ] = color;
	    ptr++;
	}
	ptr += XLINESIZE - xsize;
    }
}

void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    int src_xs, int src_ys,
    COLOR *data,
    window_manager *wm )
{
    if( !framebuffer ) return;

    int ptr = dest_y * XLINESIZE + dest_x;
    int bp = src_y * src_xs + src_x;
    for( int cy = 0; cy < dest_ys; cy++ )
    {
	for( int cx = 0; cx < dest_xs; cx++ )
	{
	    framebuffer[ ptr ] = data[ bp ];
	    ptr++;
	    bp++;
	}
	ptr += XLINESIZE - dest_xs;
	bp += src_xs - dest_xs;
    }
}

void device_redraw_framebuffer( window_manager *wm )
{
#ifdef OPENGL
    //OpenGL: 
    /* draw polygon with the screen */
    glBindTexture( GL_TEXTURE_2D, 1 );
#ifdef COLOR8BITS
    glTexImage2D( GL_TEXTURE_2D,
                  0,
                  3,
                  wm->screen_xsize, wm->screen_ysize,
                  0,
                  GL_COLOR_INDEX,
                  GL_UNSIGNED_BYTE,
                  framebuffer );
#endif
#ifdef COLOR16BITS
    glTexImage2D( GL_TEXTURE_2D,
                  0,
                  3,
                  wm->screen_xsize, wm->screen_ysize,
                  0,
                  GL_LUMINANCE,
                  GL_UNSIGNED_SHORT,
                  framebuffer );
#endif
#ifdef COLOR32BITS
    glTexImage2D( GL_TEXTURE_2D,
                  0,
                  4,
                  wm->screen_xsize, wm->screen_ysize,
                  0,
                  GL_BGRA_EXT,
                  GL_UNSIGNED_BYTE,
                  framebuffer );
#endif
    glDisable( GL_DEPTH_TEST );
    glBegin( GL_POLYGON );
    glTexCoord2f( 0, 0 );glVertex3f( -1, 1, 0 );
    glTexCoord2f( 1, 0 );glVertex3f( 1, 1, 0 );
    glTexCoord2f( 1, 1 );glVertex3f( 1, -1, 0 );
    glTexCoord2f( 0, 1 );glVertex3f( -1, -1, 0 );
    glEnd();
    glEnable( GL_DEPTH_TEST );
#ifdef WIN
    SwapBuffers( (HDC)wm->hdc );
#endif
#ifdef LINUX
    XSync( wm->dpy, 0 );
    glFlush();
    if( wm->doubleBuffer ) glXSwapBuffers( wm->dpy, wm->win );
#endif //LINUX
#endif //OPENGL

#ifdef DIRECTX
    if( g_primary_locked )
    {
	lpDDSPrimary->Unlock( g_sd.lpSurface );
	if( lpDDSPrimary )
	{
	    lpDDSPrimary->Lock( 0, &g_sd, DDLOCK_SURFACEMEMORYPTR  | DDLOCK_WAIT, 0 );
	    g_lPitch = g_sd.lPitch / COLORLEN;
	    framebuffer = (COLORPTR)g_sd.lpSurface;
	}
    }
    else
    {
	g_primary_locked = 1;
	if( lpDDSPrimary )
	{
	    lpDDSPrimary->Lock( 0, &g_sd, DDLOCK_SURFACEMEMORYPTR  | DDLOCK_WAIT, 0 );
	    g_lPitch = g_sd.lPitch / COLORLEN;
	    framebuffer = (COLORPTR)g_sd.lpSurface;
	}
    }
#endif

#ifdef PALMOS
    //PalmOS:
#ifdef PALMLOWRES
    WinDrawBitmap( (BitmapType*)g_screen_low, 0, 0 );
#else
    WinDrawBitmap( (BitmapType*)g_screen, 0, 0 );
#endif
#endif
}
