/*
    SunDog: wm_palmos.h. Platform-dependent module : PalmOS
    Copyright (C) 2002 - 2007  Zolotov Alexandr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//*** Contact info: Zolotov Alexandr (NightRadio project)
//***               Ekaterinburg. Russia.
//***               Email: nightradio@gmail.com
//***               WWW: warmplace.ru

#ifndef __WINMANAGER_PALMOS__
#define __WINMANAGER_PALMOS__

#include "time/timemanager.h"
#include <PalmOS.h>
#include <PenInputMgr.h>
#ifdef NATIVEARM
    #include "palm_functions.h"
#endif

//#################################
//## DEVICE DEPENDENT FUNCTIONS: ##
//#################################

void wait( void )
{
    int cur_ticks = time_ticks();
    int counter = 0;
    while( time_ticks() < cur_ticks + ( time_ticks_per_second() / 2 ) )
    {
	device_event_handler( 0 );
	counter++;
	if( ( counter & 0xFFFF ) == 0 ) prints( "MAIN: waiting..." );
    }
}

#ifdef FRAMEBUFFER
    BitmapType    *g_screen_low;  //Low density bitmap
    BitmapTypeV3  *g_screen;      //High density bitmap
    COLORPTR framebuffer = 0;
#endif

void device_start( char *windowname, int xsize, int ysize, int flags, window_manager *wm )
{
    //Open window:
    prints( "MAIN: device start" );

    prints( "MAIN: open form" );
    FrmGotoForm( 8888 );
    wait();

#ifdef PALMLOWRES
    xsize = 160;
    ysize = 160;
#else
    xsize = 320;
    ysize = 320;
#endif
    wm->screen_xsize = xsize;
    wm->screen_ysize = ysize;
    prints( "MAIN: device start finished" );

    //Set screen-mode and palette:
    RGBColorType pal[ 256 ];
    ulong x, y, depth;
    long a;
    x = 320; y = 320; depth = COLORBITS; 
    WinScreenMode( winScreenModeSet, &x, &y, &depth, 0 );  
#ifdef GRAYSCALE
    for( a = 0; a < 256; a++ )
    {
	pal[ a ].r = a;
	pal[ a ].g = a;
	pal[ a ].b = a;
	pal[ a ].index = a;
    }
#else
    for( a = 0; a < 256; a++ )
    {
	pal[ a ].r = ( (a<<5)&224 ) | 31;
	pal[ a ].g = ( (a<<2)&224 ) | 31;
	pal[ a ].b = (a&192)|63;
	pal[ a ].index = a;
    }
#endif
    WinPalette( winPaletteSet, 0, 256, pal );

#ifdef FRAMEBUFFER
    //Create framebuffer:
    uint16 bmp_err;
    g_screen_low = BmpCreate( wm->screen_xsize, wm->screen_xsize, COLORBITS, 0, &bmp_err );
    framebuffer = (COLORPTR)BmpGetBits( g_screen_low );
    #ifndef PALMLOWRES
	g_screen = BmpCreateBitmapV3( g_screen_low, kDensityDouble, framebuffer, 0 );
    #endif
#endif
}

int prev_move_x = -999;
int prev_move_y = -999;

void device_end( window_manager *wm )
{
    //Close window:
    prints( "MAIN: device end" );

#ifdef FRAMEBUFFER
    //Close screenbuffer:
    BmpDelete( g_screen_low );
    #ifndef PALMLOWRES
	BmpDelete( (BitmapType*)g_screen );
    #endif
#endif

    prints( "MAIN: close all forms" );
    FrmCloseAllForms();
    wait();
}

long old_chrr;

long device_event_handler( window_manager *wm )
{
    EventType event;
    UInt16 error;
    long chrr,x,y;
    char keydown = 0;
    char keyup = 0;

    EvtGetEvent( &event, 0 );

    if( wm )
    {
	keyup = 1;
	//Native ARM mode:
	UInt16 *ptr = (UInt16*)&event;
	chrr = BSwap16( ptr[4] );
	x = BSwap16( ptr[2] );
	y = BSwap16( ptr[3] );
	switch( BSwap16(event.eType) ) 
	{
#ifdef PALMLOWRES
	    case penDownEvent: send_event( 0, EVT_MOUSEBUTTONDOWN, 0, x, y, BUTTON_LEFT, 0, 1023, wm ); break;
	    case penMoveEvent: 
		if( x != prev_move_x || y != prev_move_y )
		    send_event( 0, EVT_MOUSEMOVE, 0, x, y, BUTTON_LEFT, 0, 1023, wm );
		prev_move_x = x;
		prev_move_y = y;
		break;
	    case penUpEvent:   send_event( 0, EVT_MOUSEBUTTONUP, 0, x, y, BUTTON_LEFT, 0, 1023, wm ); break;
#else
	    case penDownEvent: send_event( 0, EVT_MOUSEBUTTONDOWN, 0, x<<1, y<<1, BUTTON_LEFT, 0, 1023, wm ); break;
	    case penMoveEvent: 
		if( x != prev_move_x || y != prev_move_y )
		    send_event( 0, EVT_MOUSEMOVE, 0, x<<1, y<<1, BUTTON_LEFT, 0, 1023, wm );
		prev_move_x = x;
		prev_move_y = y;
		break;
	    case penUpEvent:   send_event( 0, EVT_MOUSEBUTTONUP, 0, x<<1, y<<1, BUTTON_LEFT, 0, 1023, wm ); break;
#endif
	    case keyDownEvent:
		keyup = 0;
	    //keyUpEvent:
	    case 0x4000:
		if( chrr >= 11 && chrr <= 12 ) keydown = 1;
		if( chrr >= 516 && chrr <= 519 ) keydown = 1;
		ulong resulted_key;
		switch( chrr )
		{
		    case 0x1E: resulted_key = KEY_UP; break;
		    case 0x1F: resulted_key = KEY_DOWN; break;
		    case 0x1C: resulted_key = KEY_LEFT; break;
		    case 0x1D: resulted_key = KEY_RIGHT; break;
		    case 0x08: resulted_key = KEY_BACKSPACE; break;
		    case 0x0A: resulted_key = KEY_ENTER; break;
		    case 0x0B: resulted_key = KEY_UP; break;
		    case 0x0C: resulted_key = KEY_DOWN; break;
		    default: resulted_key = chrr; if( resulted_key > 255 ) resulted_key = 0; break;
		}
		if( !keyup )
		    send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, 0, resulted_key, 1023, wm );
		else
		    send_event( 0, EVT_BUTTONUP, 0, 0, 0, 0, resulted_key, 1023, wm );
		break;
	}
    
	chrr = KeyCurrentState();
	if( !keydown )
	{
	    if( chrr & 0x1000000 && !( old_chrr & 0x1000000 ) ) { send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, 0, KEY_LEFT, 1023, wm ); keydown = 1; }
	    if( !( chrr & 0x1000000 ) && old_chrr & 0x1000000 ) { send_event( 0, EVT_BUTTONUP, 0, 0, 0, 0, KEY_LEFT, 1023, wm ); }
	    if( chrr & 0x2000000 && !( old_chrr & 0x2000000 ) ) { send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, 0, KEY_RIGHT, 1023, wm ); keydown = 1; }
	    if( !( chrr & 0x2000000 ) && old_chrr & 0x2000000 ) { send_event( 0, EVT_BUTTONUP, 0, 0, 0, 0, KEY_RIGHT, 1023, wm ); }
	    if( chrr & 0x4000000 && !( old_chrr & 0x4000000 ) ) { send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, 0, KEY_SPACE, 1023, wm ); keydown = 1; }
	    if( !( chrr & 0x4000000 ) && old_chrr & 0x4000000 ) { send_event( 0, EVT_BUTTONUP, 0, 0, 0, 0, KEY_SPACE, 1023, wm ); }
	}
	old_chrr = chrr;

    } //if( wm )

    if( !keydown )
    {
	if( !SysHandleEvent(&event) ) MenuHandleEvent(0,&event,&error);
    }

    if( BSwap16( event.eType ) == appStopEvent ) 
	if( wm ) send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, 0, KEY_ESCAPE, 1023, wm );

    return 0;
}

#ifdef FRAMEBUFFER

#include "wm_framebuffer.h"

#else

void device_draw_pixel( int x, int y, COLOR color, window_manager *wm )
{
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesNative );
#endif
    RGBColorType c;
    c.r = red( color );
    c.g = green( color );
    c.b = blue( color );
    WinSetForeColorRGB( &c, 0 );
    WinDrawPixel( x, y );
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesStandard );
#endif
}

COLOR device_get_pixel( int x, int y, window_manager *wm )
{
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesNative );
#endif
    RGBColorType c;
    WinGetPixelRGB( x, y, &c );
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesStandard );
#endif
    return get_color( c.r, c.g, c.b );
}

void device_draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm )
{
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesNative );
#endif
    RGBColorType c;
    c.r = red( color );
    c.g = green( color );
    c.b = blue( color );
    WinSetForeColorRGB( &c, 0 );
    WinDrawLine( x1, y1, x2, y2 );
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesStandard );
#endif
}

RectangleType box_rec;

void device_draw_box( int x, int y, int xsize, int ysize, COLOR color, window_manager *wm )
{
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesNative );
#endif
    box_rec.topLeft.x = x;
    box_rec.topLeft.y = y;
    box_rec.extent.x = xsize;
    box_rec.extent.y = ysize;
    RGBColorType c;
    c.r = red( color );
    c.g = green( color );
    c.b = blue( color );
    WinSetForeColorRGB( &c, 0 );
    WinDrawRectangle( &box_rec, 0 );
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesStandard );
#endif
}

uchar bmp[ 32 ] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 6, 5, 0, 0, 0, 0, 0 };
RectangleType bmp_rec;

void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    int src_xs, int src_ys,
    COLOR *data,
    window_manager *wm )
{
    //Create bitmap:
    bmp[ 0 ] = ( src_xs >> 8 ) & 255;
    bmp[ 1 ] = src_xs & 255;
    bmp[ 2 ] = ( src_ys >> 8 ) & 255;
    bmp[ 3 ] = src_ys & 255;
    int linesize = src_xs * COLORLEN;
    bmp[ 4 ] = ( linesize >> 8 ) & 255;
    bmp[ 5 ] = linesize & 255;
    //Type:
    bmp[ 6 ] = 0x04;
    bmp[ 7 ] = 0x00;
    bmp[ 8 ] = COLORBITS;
    bmp[ 9 ] = 2; //Version;
    BitmapTypeV3 *bitmap;
#ifdef PALMLOWRES
    bitmap = BmpCreateBitmapV3( (BitmapType*)&bmp, kDensityLow, data, 0 );
#else
    bitmap = BmpCreateBitmapV3( (BitmapType*)&bmp, kDensityDouble, data, 0 );
#endif
    //Draw it:
    bmp_rec.topLeft.x = dest_x;
    bmp_rec.topLeft.y = dest_y;
    bmp_rec.extent.x = dest_xs;
    bmp_rec.extent.y = dest_ys;
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesNative );
#endif
    WinSetClip( &bmp_rec );
    WinDrawBitmap( (BitmapType*)bitmap, dest_x - src_x, dest_y - src_y );
    bmp_rec.topLeft.x = 0;
    bmp_rec.topLeft.y = 0;
    bmp_rec.extent.x = wm->screen_xsize;
    bmp_rec.extent.y = wm->screen_ysize;
    WinSetClip( &bmp_rec );
#ifndef PALMLOWRES
    WinSetCoordinateSystem( kCoordinatesStandard );
#endif
    //Delete bitmap:
    BmpDelete( (BitmapType*)bitmap );
}

#endif

//#################################
//#################################
//#################################

#endif
