#ifndef __WINDOWMANAGER_STRUCT__
#define __WINDOWMANAGER_STRUCT__

#include "../core/core.h"
#include "../memory/memory.h"
#include "../filesystem/v3nus_fs.h"
#include "regions/device.h" //Part of the Nano-X project (functions for working with regions)

#define EVENTS 32             //must be 2/4/8/16/32...

#ifdef WIN
    #include <windows.h>
#endif

#ifdef OPENGL
    #include <GL/gl.h>
    #include <GL/glu.h>
#ifdef LINUX
    #include <GL/glx.h>
    #include <X11/keysym.h>
    #include <X11/XKBlib.h>
#endif
#endif

#ifdef X11
    #include <X11/Xlib.h>
    #include <X11/Xutil.h>
    #include <X11/keysym.h>
    #include <X11/XKBlib.h>
    #include <stdlib.h>
#endif

#ifdef DIRECTX
    #include "ddraw.h"
#endif

#ifdef LINUX 
    #include <sys/time.h>
#endif

#ifdef COLOR8BITS
    #define COLOR     uchar
    #define COLORPTR  uchar*
    #define COLORLEN  (long)1
    #define COLORBITS 8
    #define COLORMASK 0xFF
    #define CLEARCOLOR 0
#endif
#ifdef COLOR16BITS
    #define COLOR     uint16
    #define COLORPTR  uint16*
    #define COLORLEN  (long)2
    #define COLORBITS 16
    #define COLORMASK 0xFFFF
    #define CLEARCOLOR 0
#endif
#ifdef COLOR32BITS
    #define COLOR     ulong
    #define COLORPTR  ulong*
    #define COLORLEN  (long)4
    #define COLORBITS 32
    #define COLORMASK 0xFFFFFFFF
    #define CLEARCOLOR 0xFF000000
#endif

#define CW COLORMASK
#define CB CLEARCOLOR

#define WIN_INIT_FLAG_SCALABLE	    1

enum {
    EVT_NULL = 0,
    EVT_GETDATASIZE,
    EVT_AFTERCREATE,
    EVT_BEFORECLOSE,
    EVT_DRAW,
    EVT_SCREENRESIZE,	    //After main screen resize
    EVT_UNFOCUS,	    //When user click on other window
    EVT_MOUSEBUTTONDOWN,
    EVT_MOUSEBUTTONUP,
    EVT_BUTTONDOWN,
    EVT_BUTTONUP,
    EVT_MOUSEMOVE
};

//buttons:
#define BUTTON_LEFT 1
#define BUTTON_MIDDLE 2
#define BUTTON_RIGHT 4
#define BUTTON_SCROLLUP 8
#define BUTTON_SCROLLDOWN 16

//Keyboard codes:
//There is full ASCII table with 256 key-codes
//Some of them:
#define KEY_SPACE       0x20
#define KEY_ENTER       10
#define KEY_BACKSPACE   127
#define KEY_TAB         9
#define KEY_ESCAPE      27

//Additional keys:
enum {
    KEY_F1 = 256,
    KEY_F2,
    KEY_F3,
    KEY_F4,
    KEY_F5,
    KEY_F6,
    KEY_F7,
    KEY_F8,
    KEY_UP,
    KEY_DOWN,
    KEY_LEFT,
    KEY_RIGHT,
    KEY_INSERT,
    KEY_DELETE,
    KEY_HOME,
    KEY_END,
    KEY_PAGEUP,
    KEY_PAGEDOWN,
    KEY_CAPS
};

//Special key flags:
#define KEY_SHIFT       512
#define KEY_CTRL        1024
#define KEY_ALT         2048

//Scroll flags:
#define SF_LEFT         1
#define SF_RIGHT        2
#define SF_UP           4
#define SF_DOWN         8
#define SF_STATICWIN    16

#define WINDOWPTR	sundog_window*
#define WINDOW		sundog_window

#define DECOR_BORDER_SIZE   3
#define DECOR_HEADER_SIZE   14

#define SCROLLBAR_SIZE	    18

#define WIN_MAX_COMS	    8
#define CEND		    40000
#define CWIN		    30000
#define CX1		    30001
#define CX2		    30002
#define CY1		    30003
#define CY2		    30004
#define CSUB		    30005
#define CADD		    30006
#define CPERC		    30007
#define CMAXVAL		    30008
#define CMINVAL		    30009

struct sundog_window
{
    int		    visible;
    char	    *name;			    //window name
    int		    x, y;			    //x/y (relative)
    int		    screen_x, screen_y;		    //real x/y
    int		    xsize, ysize;		    //window size
    COLOR	    color;
    MWCLIPREGION    *reg;
    WINDOWPTR	    parent;
    WINDOWPTR	    *childs;
    int		    childs_num;
    int		    (*win_handler)( void*, void* ); //window handler: (*event, *window_manager)
    void	    *data;

    int		    x1com[ WIN_MAX_COMS ];	    //Controllers of window coordinates (special commands)
    int		    y1com[ WIN_MAX_COMS ];
    int		    x2com[ WIN_MAX_COMS ];
    int		    y2com[ WIN_MAX_COMS ];
    int		    controllers_defined;
    int		    controllers_calculated;

    int		    (*action_handler)( void*, void*, void* );
    int		    action_result;
    void	    *handler_data;

    //Direct Draw properties:
    COLORPTR	    ddraw_buf;
    int		    ddraw_xsize;
    int		    ddraw_ysize;
};

struct sundog_event
{
    long            event_type;		//event type
    WINDOWPTR	    event_win;
    int		    all_childs;		//1 - send this event to each childred in any case
    long            x;
    long            y;
    long            mouse_button;
    long            button;
    long            pressure;		//Key pressure (0..1023)
};

struct sundog_font
{
    uchar	    *data;
    char	    xsize[ 256 ];
    char	    real_width[ 256 ];
    char	    ysize[ 256 ];
    char	    y_offset[ 256 ];
    int		    descent;
    char	    max_ysize;
    int		    offset[ 256 ];
    COLOR	    rendered[ 32 * 32 ];
};

struct window_manager
{
    long            events_count;	    //number of events to execute
    long	    current_event_num;
    sundog_event    events[ EVENTS ];

    WINDOWPTR	    *trash;

    WINDOWPTR	    root_win;

    int             screen_xsize;
    int             screen_ysize;

    sundog_font	    *user_font;

    COLOR	    colors[ 16 ];	    //System palette
    COLOR	    white;
    COLOR	    black;
    					  
    long            wm_initialized;
    
    long	    exit_request;

    //Direct Draw to some buffer created in some window: ======================================================
    COLOR           *g_screen;		    //graphics screen pointer (each element is BGR233 / RGB565 / RGB888 )
    WINDOWPTR	    cur_window;		  
    long            pscreen_x;		    //screen size (x) in pixels
    long            pscreen_y;		    //screen size (y) in pixels
    COLOR           cur_font_color;
    bool            cur_font_invert;
    bool            cur_font_draw_bgcolor; 
    COLOR           cur_font_bgcolor;
    int		    cur_transparency;
    //=========================================================================================================

    int		    pen_x;
    int		    pen_y;
    WINDOWPTR	    focus_win;
    WINDOWPTR	    prev_focus_win;

    char	    temp_string[ 512 ];

    uint16          buttons_table[256];	    //DEVICE DEPENDENT: keyboard table for event-manager
    ulong           hdc;		    //DEVICE DEPENDENT: graphics content handler (for win32)
#ifdef OPENGL
#ifdef LINUX
    GLboolean       doubleBuffer;
#endif //LINUX
    int		    real_window_width;
    int		    real_window_height;
#endif //OPENGL
#ifdef DIRECTX
    LPDIRECTDRAWSURFACE lpDDSPrimary;	    //DEVICE DEPENDENT: DirectDraw primary surface (for win32)
    LPDIRECTDRAWSURFACE lpDDSBack;	    //DEVICE DEPENDENT: DirectDraw back surface (for win32)
#endif
#ifdef GDI
    ulong           gdi_bitmap_info[2048];  //DEVICE DEPENDENT: GDI bitmap info (for win32)
#endif
#ifdef WIN
    HINSTANCE       hCurrentInst;
    HINSTANCE       hPreviousInst; 
    LPSTR           lpszCmdLine;
    int             nCmdShow;
#endif
#ifdef LINUX
    int             argc;
    char            **argv;
#ifndef TEXTMODE
    Display         *dpy;
    Colormap	    cmap;
    Window          win;
    Visual          *win_visual;
    GC              win_gc;
    XImage          *win_img;
    XImage          *win_img_back_pattern;
    char            *win_img_back_pattern_data;
    long	    win_img_depth;
    char            *win_buffer;
    int	            win_depth;
#endif //TEXTMODE
#endif //LINUX
};

#endif

