#ifndef __WINDOWMANAGER__
#define __WINDOWMANAGER__

#include "struct.h"
#include "../utils/utils.h"

//################################
//## MAIN FUNCTIONS:            ##
//################################

void win_init( char*, int xsize, int ysize, int flags, window_manager* );     
                                                  //init window manager
void win_close( window_manager* );	          //close window manager

//################################
//## WINDOWS FUNCTIONS:         ##
//################################

WINDOWPTR new_window( 
    char *name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color,
    WINDOWPTR parent, 
    int (*)( sundog_event*, window_manager* ),
    window_manager *wm );
void set_window_controller( WINDOWPTR win, int ctrl_num, window_manager *wm, ... );
void remove_window( WINDOWPTR win, window_manager *wm );
void set_handler( WINDOWPTR win, int (*handler)(void*,WINDOWPTR,window_manager*), void *handler_data, window_manager *wm );
void draw_window( WINDOWPTR win, window_manager *wm );
void show_window( WINDOWPTR win, window_manager *wm );
void hide_window( WINDOWPTR win, window_manager *wm );
void recalc_regions( window_manager *wm );
void send_event( WINDOWPTR win, int event_type, int all_childs, int x, int y, int mouse_button, int button, int pressure, window_manager* );
                                                  //send event to some window
int handle_event_by_window( sundog_event *evt, window_manager *wm );

//################################
//## WINDOWS DECORATIONS:       ##
//################################

WINDOWPTR new_window_with_decorator( 
    char *name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color,
    WINDOWPTR parent, 
    int (*)( sundog_event*, window_manager* ),
    window_manager *wm );

//################################
//## DRAWING FUNCTIONS:         ##
//################################

inline COLOR get_color( uchar r, uchar g, uchar b ); //Get color value by RGB
inline long red( COLOR c );
inline long green( COLOR c );
inline long blue( COLOR c );
inline COLOR blend( COLOR c1, COLOR c2, int value );
int char_x_size( uchar c, window_manager *wm );
int char_y_size( window_manager *wm );
int string_size( char *str, window_manager *wm );

void win_draw_box( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager *wm );
void win_draw_frame( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager *wm );
void win_draw_frame3d( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, int inout, window_manager *wm );
void win_draw_bitmap( 
    WINDOWPTR win, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR *data, 
    window_manager *wm );
void win_draw_char( WINDOWPTR win, int c, int x, int y, COLOR f, COLOR b, window_manager *wm );
void win_draw_string( WINDOWPTR win, char *str, int x, int y, COLOR f, COLOR b, window_manager *wm );
void win_draw_line( WINDOWPTR win, int x1, int y1, int x2, int y2, COLOR color, window_manager *wm );

extern uint16 *g_mul;

inline COLOR get_color( uchar r, uchar g, uchar b )
{
    COLOR res = 0; //resulted color
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = (COLOR)( (int)( ( r + g + b + b ) >> 2 ) );
    #else
	res += ( b & 192 );
	res += ( g & 224 ) >> 2;
	res += ( r >> 5 );
    #endif
#endif
#ifdef COLOR16BITS
    res += ( r >> 3 ) << 11;
    res += ( g >> 2 ) << 5;
    res += ( b >> 3 );
    #ifndef NONPALM
    res = ( res >> 8 ) + ( ( res & 255 ) << 8 );
    #endif
#endif
#ifdef COLOR32BITS
    res += 255 << 24;
    res += r << 16;
    res += g << 8;
    res += b;
#endif
    return res;
}

inline long red( COLOR c )
{
    long res = 0;
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = ( c << 5 ) & 255;
    #endif
#endif
#ifdef COLOR16BITS
    #ifndef NONPALM
    c = ( c >> 8 ) + ( ( c & 255 ) << 8 );
    #endif
    res = ( ( c >> 11 ) << 3 ) & 0xF8;
#endif
#ifdef COLOR32BITS
    res = ( c >> 16 ) & 255;
#endif
    return res;
}

inline long green( COLOR c )
{
    long res = 0;
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = ( c << 2 ) & 0xE0;
    #endif
#endif
#ifdef COLOR16BITS
    #ifndef NONPALM
    c = ( c >> 8 ) + ( ( c & 255 ) << 8 );
    #endif
    res = ( ( c >> 5 ) << 2 ) & 0xFC;
#endif
#ifdef COLOR32BITS
    res = ( c >> 8 ) & 255;
#endif
    return res;
}

inline long blue( COLOR c )
{
    long res = 0;
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = c & 192;
    #endif
#endif
#ifdef COLOR16BITS
    #ifndef NONPALM
    c = ( c >> 8 ) + ( ( c & 255 ) << 8 );
    #endif
    res = ( c << 3 ) & 0xF8;
#endif
#ifdef COLOR32BITS
    res = c & 255;
#endif
    return res;
}

inline COLOR blend( COLOR c1, COLOR c2, int value )
{
    if( value == 0 ) return c1;
    else
    if( value > 255 ) return c2;
#ifdef COLOR32BITS
    int r1 = ( c1 >> 16 ) & 255;
    int g1 = ( c1 >> 8 ) & 255;
    int b1 = c1 & 255;
    int r2 = ( c2 >> 16 ) & 255;
    int g2 = ( c2 >> 8 ) & 255;
    int b2 = c2 & 255;
    int ival = 256 - value; 
    ival <<= 8;
    value <<= 8;
    int r3 = ( g_mul[ r1 | ival ] + g_mul[ r2 | value ] ) >> 8;
    int g3 = ( g_mul[ g1 | ival ] + g_mul[ g2 | value ] ) >> 8;
    int b3 = ( g_mul[ b1 | ival ] + g_mul[ b2 | value ] ) >> 8;
    return ( r3 << 16 ) | ( g3 << 8 ) | b3;
#endif
#ifdef COLOR16BITS
    #ifdef PALMOS
	c1 = ( c1 >> 8 ) + ( ( c1 & 255 ) << 8 );
	c2 = ( c2 >> 8 ) + ( ( c2 & 255 ) << 8 );
    #endif
    int r1 = c1 >> 10; r1 |= 1;
    int g1 = ( c1 >> 5 ) & 63; g1 |= 1;
    int b1 = c1 & 31; b1 <<= 1; b1 |= 1;
    int r2 = c2 >> 10; r2 |= 1;
    int g2 = ( c2 >> 5 ) & 63; g2 |= 1;
    int b2 = c2 & 31; b2 <<= 1; b2 |= 1;
    int ival = 256 - value; 
    ival <<= 4; ival &= ( 63 << 6 );
    value <<= 4; value &= ( 63 << 6 );
    int r3 = g_mul[ r1 | ival ] + g_mul[ r2 | value ];
    int g3 = g_mul[ g1 | ival ] + g_mul[ g2 | value ];
    int b3 = g_mul[ b1 | ival ] + g_mul[ b2 | value ];
    r3 >>= 7;
    g3 >>= 6;
    b3 >>= 7;
    int res = ( r3 << 11 ) | ( g3 << 5 ) | b3;
    #ifdef PALMOS
	return ( res >> 8 ) + ( ( res & 255 ) << 8 );
    #else    
	return res;
    #endif
#endif
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	int ival = 256 - value; 
	ival <<= 8;
	value <<= 8;
	int res = ( g_mul[ c1 | ival ] + g_mul[ c2 | value ] ) >> 8;
	return (COLOR)res;
    #else
	int b1 = c1 >> 5; b1 &= ~1;
	int g1 = ( c1 >> 3 ) & 7; 
	int r1 = c1 & 7; 
	int b2 = c2 >> 5; b2 &= ~1;
	int g2 = ( c2 >> 3 ) & 7; 
	int r2 = c2 & 7; 
	int ival = 256 - value; 
	ival <<= 3;
	value <<= 3;
	int r3 = g_mul[ r1 | ival ] + g_mul[ r2 | value ];
	int g3 = g_mul[ g1 | ival ] + g_mul[ g2 | value ];
	int b3 = g_mul[ b1 | ival ] + g_mul[ b2 | value ];
	b3 >>= 9;
	g3 >>= 8;
	r3 >>= 8;
	return ( b3 << 6 ) | ( g3 << 3 ) | r3;
    #endif
#endif
    return 0;
}

//################################
//## FONTS:                     ##
//################################

sundog_font *win_load_font( char *filename, window_manager *wm );
void win_remove_font( sundog_font *fnt );

//################################
//## DIRECT DRAW FUNCTIONS:     ##
//################################

void direct_draw_init( WINDOWPTR win, int xsize, int ysize, window_manager *wm );
void direct_draw_deinit( WINDOWPTR win, window_manager *wm );
void direct_draw_set_current_window( WINDOWPTR win, window_manager *wm );
void direct_draw_resize( WINDOWPTR win, int new_xsize, int new_ysize, window_manager *wm );
void direct_draw( WINDOWPTR win, window_manager *wm );

void draw_screen( window_manager *wm );
void clear_screen( window_manager *wm );
void draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm );
void draw_frame( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm );
void draw_box( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm );
int draw_char( uchar c, int x, int y, window_manager *wm );
void draw_string( char *str, int x, int y, window_manager *wm );

//################################
//## SATNDART WINDOW HANDLERS:  ##
//################################

int desktop_handler( sundog_event *evt, window_manager *wm );
int label_handler( sundog_event *evt, window_manager *wm );
int text_handler( sundog_event *evt, window_manager *wm );
void text_set_text( WINDOWPTR win, char *text, window_manager *wm );
char *text_get_text( WINDOWPTR win, window_manager *wm );
extern COLORPTR CREATE_BITMAP_BUTTON;
extern int CREATE_BITMAP_BUTTON_XSIZE;
extern int CREATE_BITMAP_BUTTON_YSIZE;
extern char *CREATE_BUTTON_WITH_TEXT_MENU;
int button_handler( sundog_event *evt, window_manager *wm );
int list_handler( sundog_event *evt, window_manager *wm );
list_data *list_get_data( WINDOWPTR win, window_manager *wm );
extern int CREATE_VERTICAL_SCROLLBAR;
extern int CREATE_REVERSE_SCROLLBAR;
int scrollbar_handler( sundog_event *evt, window_manager *wm );
void scrollbar_set_parameters( WINDOWPTR win, int cur, int max_value, int page_size, int step_size, window_manager *wm );
void scrollbar_set_value( WINDOWPTR win, int val, window_manager *wm );
int scrollbar_get_value( WINDOWPTR win, window_manager *wm );
extern char *FILES_PROPS;
extern char *FILES_MASK;
extern char *FILES_RESULTED_FILENAME;
int files_handler( sundog_event *evt, window_manager *wm );
extern char *DIALOG_OK_TEXT;
extern char *DIALOG_CANCEL_TEXT;
extern char *DIALOG_TEXT;
int dialog_handler( sundog_event *evt, window_manager *wm );
extern char *CREATE_TEXT_POPUP;
int popup_handler( sundog_event *evt, window_manager *wm );
int popup_menu( char* name, char* items, int x, int y, COLOR color, window_manager *wm );

//################################
//## DIALOGS:                   ##
//################################

char *dialog_open_file( char *name, char *mask, char *id, window_manager *wm );
int dialog( char *name, char *ok, char *cancel, window_manager *wm );

//################################
//## DEVICE DEPENDENT:          ##
//################################

long device_event_handler( window_manager* );     //device event handler (return 1 for EXIT)
void device_start( char*, int xsize, int ysize, int flags, window_manager* ); 
                                                  //device start (before main loop)
void device_end( window_manager* );	          //device end   (after main loop)

void device_draw_pixel( int x, int y, COLOR color, window_manager *wm );
COLOR device_get_pixel( int x, int y, window_manager *wm );
void device_draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm );
void device_draw_box( int x, int y, int xsize, int ysize, COLOR color, window_manager *wm );
void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    int src_xs, int src_ys,
    COLOR *data,
    window_manager *wm );
void device_redraw_framebuffer( window_manager *wm );

//################################
//## MACROSES:                  ##
//################################

#define FIRST_PART_OF_THE_EVENT_LOOP( wmanager ) \
    sundog_event current_event; \
    device_event_handler( wmanager ); \
    if( wmanager->events_count ) \
    { \
	/*There are unhandled events:*/ \
	/*Copy current event to "current_event" buffer (prepare it for handling):*/ \
	mem_copy( &current_event, &wmanager->events[ wmanager->current_event_num ], sizeof( sundog_event ) ); \
	/*This event will be handled. So decrement count of events:*/ \
	wmanager->events_count--; \
	/*And increment FIFO pointer:*/ \
	wmanager->current_event_num = ( wmanager->current_event_num + 1 ) & ( EVENTS - 1 ); \
	/*Now we can handle our event:*/ \
	if( !user_event_handler( &current_event, wmanager ) || current_event.all_childs ) \
	{ \
	    /*Event hot handled by simple event handler.*/ \
	    /*Sent it to windows:*/ \
	    handle_event_by_window( &current_event, wmanager ); \
	} \
	/*HANDLE EVENT BY USER*/

#define LAST_PART_OF_THE_EVENT_LOOP( wmanager ) \
    g_frame++; \
    if( g_frame % 4 == 0 || !g_skip_frames ) \
    { \
	user_screen_redraw( wmanager ); \
	device_redraw_framebuffer( wmanager ); \
    } \
    if( wmanager->exit_request ) break; 

#endif

