#ifndef __SIMPLEXML_HEADER__
#define __SIMPLEXML_HEADER__

#define NUMBER_OF_CHARS 94 //ASCII
#define FIRST_CHAR 32
#define NODES_PER_PAGE 16
#define CHILDS_PER_PAGE 16

/*
It's simple XML library. 
Object = tag with name and type (4 bytes string).
Each object can contain own value.
For example there are two empty objects:
<SMPL name="HELLO"\>
<SMPL name="HELP"\>
XML Node - array of links to the next characters in XML object name. It's base of our fast string searching algorythm.
Example:
node0
  |
node1 (H)
  |
node2 (E)
  |
node3 (L)
  |\
  | \
  |  \
  | node6 (P)
node4 (L)
  |
node5 (O)
In this example our XML tree contents two words (objects): HELLO and HELP.
Flag end_of_word is toggle in a nodes 5 and 6.
Node with toggled end_of_word IS object.
Objects can has parents and childs. 
Use the "." character for switching to a next level of hierarchy: PARENT.CHILD.CHILD2...
*/

struct data_header
{
    int xsize;
    int ysize;
};

struct xml_node
{
    int links[ NUMBER_OF_CHARS ]; //Numbers of next nodes for each character
    bool end_of_word;
    bool busy;
    int value;      //Object's value
    unsigned int type;//Object's type
    char *data;     //Object's binary data
    char *name;     //Object's name
    int *childs;    //Object's childs
    int childs_num; //Number of childs
};

struct xml_data
{
    char *xt;     //XML tree
    int xt_size;  //Size of XML tree (bytes)
};

//################################
//## MAIN FUNCTIONS:            ##
//################################

void xml_init( xml_data *x ) sec2;
void xml_close( xml_data *x ) sec2;

int xml_get_value( int node, xml_data *x ); //Return 0 if error
int xml_get_type( int node, xml_data *x ); //Return 0 if error
void xml_set_value( int node, int value, unsigned int type, xml_data *x );
void xml_create_data( int node, int xsize, int ysize, xml_data *x );
void xml_resize_data( int node, int xsize, int ysize, xml_data *x );
data_header *xml_get_data_header( int node, xml_data *x );
void *xml_get_data_ptr( int node, xml_data *x );
void xml_delete_data( int node, xml_data *x );

xml_node *xml_get_node_ptr( int node, xml_data *x );
int xml_create_node( xml_data *x );

int xml_find_object( int parent_node, char *name, xml_data *x ); //Find object (node number). If it's not exists, create it.

/*
void xml_load( char *filename, xml_data *x );
void xml_save( char *filename, char *description, xml_data *x );
*/

#endif /* __SIMPLEXML_HEADER__ */
